import sys
from typing import (
    Any,
    ClassVar,
    Iterator,
    List,
    Match,
    Optional,
    Pattern,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from mistune._types import State, Tokens
from mistune.inline_parser import InlineParser, RendererT
from mistune.renderers import DataT
from mistune.scanner import Matcher, ScannerParser
from typing_extensions import Literal, NotRequired, Required, TypeAlias, TypedDict

ParsedBlockType = Literal[
    # base block parsers
    "heading",
    "newline",
    "thematic_break",
    "block_code",
    "block_html",
    "block_quote",
    "block_text",
    "list",
    "list_item",
    "paragraph",
    # plugin 'table'
    "table",
    "table_head",
    "table_body",
    "table_cell",
    "table_row",
    # plugin 'url'
    "url",
]
_ParsedBlock: TypeAlias = "ParsedBlock"
if sys.version_info >= (3, 7):
    ParsedTypeT = TypeVar("ParsedTypeT")
    ParsedBlock = TypedDict(
        "ParsedBlock",
        {
            "type": Required[ParsedTypeT],
            "blank": NotRequired[bool],
            "raw": NotRequired[str],
            "text": NotRequired[str],
            "params": NotRequired[Tuple[Union[int, str], ...]],
            "children": NotRequired[List[_ParsedBlock]],
        },
        total=False,
    )
    ParsedBlockHeading = ParsedBlock[Literal["heading"]]
    ParsedBlockNewline = ParsedBlock[Literal["newline"]]
    ParsedBlockThematicBreak = ParsedBlock[Literal["thematic_break"]]
    ParsedBlockBlockCode = ParsedBlock[Literal["block_code"]]
    ParsedBlockBlockHTML = ParsedBlock[Literal["block_html"]]
    ParsedBlockBlockQuote = ParsedBlock[Literal["block_quote"]]
    ParsedBlockBlockText = ParsedBlock[Literal["block_text"]]
    ParsedBlockList = ParsedBlock[Literal["list"]]
    ParsedBlockListItem = ParsedBlock[Literal["list_item"]]
    ParsedBlockParagraph = ParsedBlock[Literal["paragraph"]]
else:  # python 3.6  # no TypedDict+Generic support with TypeVar
    ParsedBlock = TypedDict(
        "ParsedBlock",
        {
            # best we can do is define an 'AnyOf' allowed literals
            # we cannot provide explicitly which literal is returned each time
            "type": Required[ParsedBlockType],
            "blank": NotRequired[bool],
            "raw": NotRequired[str],
            "text": NotRequired[str],
            "params": NotRequired[Tuple[Union[int, str], ...]],
            "children": NotRequired[List[_ParsedBlock]],
        },
        total=False,
    )
    ParsedBlockHeading = ParsedBlock
    ParsedBlockNewline = ParsedBlock
    ParsedBlockThematicBreak = ParsedBlock
    ParsedBlockBlockCode = ParsedBlock
    ParsedBlockBlockHTML = ParsedBlock
    ParsedBlockBlockQuote = ParsedBlock
    ParsedBlockBlockText = ParsedBlock
    ParsedBlockList = ParsedBlock
    ParsedBlockListItem = ParsedBlock
    ParsedBlockParagraph = ParsedBlock

class BlockParser(ScannerParser):
    NEWLINE: Pattern[str]
    DEF_LINK: Pattern[str]
    AXT_HEADING: Pattern[str]
    SETEX_HEADING: Pattern[str]
    THEMATIC_BREAK: Pattern[str]
    INDENT_CODE: Pattern[str]
    FENCED_CODE: Pattern[str]
    BLOCK_QUOTE: Pattern[str]
    LIST_START: Pattern[str]
    BLOCK_HTML: Pattern[str]
    LIST_MAX_DEPTH: ClassVar[int]
    BLOCK_QUOTE_MAX_DEPTH: ClassVar[int]

    scanner_cls: ClassVar[Type[Matcher]] = Matcher
    block_quote_rules: List[str]
    list_rules = List[str]

    def __init__(self) -> None: ...
    def parse_newline(self, m: Match[str], state: State) -> ParsedBlockNewline: ...
    def parse_thematic_break(self, m: Match[str], state: State) -> ParsedBlockThematicBreak: ...
    def parse_indent_code(self, m: Match[str], state: State) -> ParsedBlockBlockCode: ...
    def parse_fenced_code(self, m: Match[str], state: State) -> ParsedBlockBlockCode: ...
    def tokenize_block_code(
        self, code: str, info: Tuple[str, ...], state: State
    ) -> ParsedBlockBlockCode: ...
    def parse_axt_heading(self, m: Match[str], state: State) -> ParsedBlockHeading: ...
    def parse_setex_heading(self, m: Match[str], state: State) -> ParsedBlockHeading: ...
    def tokenize_heading(self, text: str, level: int, state: State) -> ParsedBlockHeading: ...
    def get_block_quote_rules(self, depth: int) -> List[str]: ...
    def parse_block_quote(self, m: Match[str], state: State) -> ParsedBlockBlockQuote: ...
    def get_list_rules(self, depth: int) -> List[str]: ...
    def parse_list_start(
        self, m: Match[str], state: State, string: str
    ) -> Tuple[ParsedBlockList, int]: ...
    def parse_list_item(
        self, text: str, depth: int, state: State, rules: List[str]
    ) -> ParsedBlockListItem: ...
    def normalize_list_item_text(self, text: str) -> str: ...
    def parse_block_html(self, m: Match[str], state: State) -> ParsedBlockBlockHTML: ...
    def parse_def_link(self, m: Match[str], state: State) -> None: ...
    def parse_text(
        self, text: str, state: State
    ) -> Union[ParsedBlockBlockText, List[ParsedBlockParagraph]]: ...
    def parse(self, s: str, state: State, rules: Optional[List[str]] = None) -> Tokens: ...
    def render(
        self, tokens: Tokens, inline: InlineParser[RendererT], state: State
    ) -> Any: ...  # technically DataT, but defined by 'inline.renderer.finalize'
    def _iter_render(
        self, tokens: Tokens, inline: InlineParser[RendererT], state: State
    ) -> Iterator[DataT]: ...

def cleanup_lines(s: str) -> str: ...
def expand_leading_tab(text: str) -> str: ...



class BaseCache:
    def get(self, key: str) -> bytes | None: ...
    def set(self, key: str, value: bytes, expires: int | None = None) -> None: ...
    def delete(self, key: str) -> None: ...
    def close(self) -> None: ...



import inspect
import os
import pathlib
from pathlib import Path
from typing import Any, Dict, List, Optional, cast

from rdflib import Graph
from rdflib.compare import to_isomorphic
from requests import Session

import schema_salad.metaschema as cg_metaschema
from schema_salad import codegen
from schema_salad.avro.schema import Names
from schema_salad.fetcher import DefaultFetcher
from schema_salad.python_codegen import PythonCodeGen
from schema_salad.python_codegen_support import LoadingOptions
from schema_salad.schema import load_schema

from .util import basket_file_uri, cwl_file_uri, get_data, metaschema_file_uri


def test_safe_identifiers() -> None:
    """Affirm correct construction of identifiers safe for Python."""
    assert PythonCodeGen.safe_name("5.8s_pattern") == "_5_8s_pattern"


def test_cwl_gen(tmp_path: Path) -> None:
    src_target = tmp_path / "src.py"
    python_codegen(cwl_file_uri, src_target)
    assert os.path.exists(src_target)
    with open(src_target) as f:
        assert "class Workflow(Process)" in f.read()


def test_meta_schema_gen(tmp_path: Path) -> None:
    src_target = tmp_path / "src.py"
    python_codegen(metaschema_file_uri, src_target)
    assert os.path.exists(src_target)
    with open(src_target) as f:
        assert "class RecordSchema(Saveable):" in f.read()


def test_meta_schema_gen_up_to_date(tmp_path: Path) -> None:
    src_target = tmp_path / "src.py"
    python_codegen(metaschema_file_uri, src_target)
    assert os.path.exists(src_target)
    with open(src_target) as f:
        assert f.read() == inspect.getsource(cg_metaschema)


def test_meta_schema_gen_no_base(tmp_path: Path) -> None:
    src_target = tmp_path / "src.py"
    python_codegen(basket_file_uri, src_target)
    assert os.path.exists(src_target)
    with open(src_target) as f:
        assert "class Basket" in f.read()


def python_codegen(
    file_uri: str,
    target: Path,
    parser_info: Optional[str] = None,
    package: Optional[str] = None,
) -> None:
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(file_uri)
    assert isinstance(avsc_names, Names)
    schema_raw_doc = metaschema_loader.fetch(file_uri)
    schema_doc, schema_metadata = metaschema_loader.resolve_all(schema_raw_doc, file_uri)
    codegen.codegen(
        "python",
        cast(List[Dict[str, Any]], schema_doc),
        schema_metadata,
        document_loader,
        target=str(target),
        parser_info=parser_info,
        package=package,
    )


def test_default_parser_info(tmp_path: Path) -> None:
    src_target = tmp_path / "src.py"
    python_codegen(metaschema_file_uri, src_target)
    assert os.path.exists(src_target)
    with open(src_target) as f:
        assert 'def parser_info() -> str:\n    return "org.w3id.cwl.salad"' in f.read()


def test_parser_info(tmp_path: Path) -> None:
    src_target = tmp_path / "src.py"
    python_codegen(metaschema_file_uri, src_target, parser_info="cwl")
    assert os.path.exists(src_target)
    with open(src_target) as f:
        assert 'def parser_info() -> str:\n    return "cwl"' in f.read()


def test_use_of_package_for_parser_info(tmp_path: Path) -> None:
    src_target = tmp_path / "src.py"
    python_codegen(metaschema_file_uri, src_target, package="cwl")
    assert os.path.exists(src_target)
    with open(src_target) as f:
        assert 'def parser_info() -> str:\n    return "cwl"' in f.read()


def test_graph_property() -> None:
    """Test the RDFLib Graph representation of the `$schemas` directive."""
    schema = cast(str, get_data("tests/EDAM.owl"))
    fetcher = DefaultFetcher({}, Session())
    fetchurl = pathlib.Path(schema).resolve().as_uri()
    content = fetcher.fetch_text(fetchurl)
    graph = Graph()
    graph.parse(data=content, format="xml", publicID=fetchurl)
    loading_options = LoadingOptions(schemas=[schema])
    assert to_isomorphic(graph) == to_isomorphic(loading_options.graph)


def test_graph_property_cache() -> None:
    """Test that LoadingOptions properly cache the `$schemas` RDFLib Graph representations."""
    schema = cast(str, get_data("tests/EDAM.owl"))
    loading_options = LoadingOptions(schemas=[schema])
    graph1 = loading_options.graph
    graph2 = loading_options.graph
    assert graph1 == graph2


def test_graph_property_empty_schema() -> None:
    """Test that an empty RDFLib Graph is returned when not `$schemas` directive is present."""
    loading_options = LoadingOptions()
    assert to_isomorphic(loading_options.graph) == to_isomorphic(Graph())



from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterable,
    Iterator,
    List,
    Match,
    Pattern,
    Tuple,
    Type,
    Union,
)

from mistune._types import State
from typing_extensions import TypeAlias

MethodFunc: TypeAlias = Callable[["ScannerParser", Match, State], Any]
RuleMethod = Tuple[Pattern[str], MethodFunc]
Lexicon = List[Tuple[Pattern[str], Tuple[str, RuleMethod]]]
TextParser = Callable[[str, State], str]

class SREScanner:  # real type unknown
    def __init__(self, lexicon: Lexicon) -> None: ...
    def scanner(self, string: str) -> Pattern[str]: ...

class Scanner(object):
    scanner: ClassVar[SREScanner]
    lexicon: Lexicon

    def iter(
        self,
        string: str,
        state: State,
        parse_text: TextParser,
    ) -> Iterator[str]: ...

class Matcher:
    PARAGRAPH_END: ClassVar[Pattern[str]]
    lexicon: Lexicon

    def __init__(self, lexicon: Lexicon) -> None: ...
    def search_pos(self, string: str, pos: int) -> int: ...
    def iter(self, string: str, state: State, parse_text: TextParser) -> Iterator[str]: ...

class ScannerParser:
    RULE_NAMES: ClassVar[Tuple[str, ...]]
    scanner_cls: ClassVar[Union[Type[SREScanner], Type[Matcher]]]
    rules: List[str]
    rule_methods: Dict[str, RuleMethod]
    _cached_sc: Dict[str, Scanner]

    def __init__(self) -> None: ...
    def register_rule(self, name: str, patter: Pattern[str], method: RuleMethod) -> None: ...
    def get_rule_pattern(self, name: str) -> Pattern[str]: ...
    def get_rule_method(self, name: str) -> RuleMethod: ...
    def parse_text(
        self, text: str, state: State
    ) -> Any: ...  # abstract method, return specialized by implementation
    def _scan(self, s: str, state: State, rules: Iterable[str]) -> Iterator[str]: ...
    def _create_scanner(self, rules: Iterable[str]) -> Scanner: ...



from typing import Dict, Union

from rdflib.graph import ConjunctiveGraph, Graph

Stats = Dict[str, Union[int, str]]

class IsomorphicGraph(ConjunctiveGraph):
    pass

def to_isomorphic(graph: Graph = ...) -> IsomorphicGraph: ...



import pathlib
from typing import (
    IO,
    Any,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    overload,
)

from rdflib import query
from rdflib.collection import Collection
from rdflib.paths import Path
from rdflib.resource import Resource
from rdflib.term import BNode, Identifier, Node

class Graph(Node):
    base: Any = ...
    context_aware: bool = ...
    formula_aware: bool = ...
    default_union: bool = ...
    def __init__(
        self,
        store: str = ...,
        identifier: Optional[Any] = ...,
        namespace_manager: Optional[Any] = ...,
        base: Optional[Any] = ...,
    ) -> None: ...
    store: Any = ...
    identifier: Any = ...
    namespace_manager: Any = ...
    def toPython(self) -> str: ...
    def destroy(self, configuration: Any) -> None: ...
    def commit(self) -> None: ...
    def rollback(self) -> None: ...
    def open(self, configuration: Any, create: bool = ...) -> Any: ...
    def close(self, commit_pending_transaction: bool = ...) -> None: ...
    def add(self, triple: Any) -> None: ...
    def addN(self, quads: Any) -> None: ...
    def remove(self, triple: Any) -> None: ...
    def triples(
        self,
        triple: Tuple[
            Optional[Union[str, Identifier]],
            Optional[Union[str, Identifier]],
            Optional[Identifier],
        ],
    ) -> Iterator[Tuple[Identifier, Identifier, Identifier]]: ...
    def __getitem__(
        self, item: slice | Path | Node
    ) -> Iterator[
        Tuple[Identifier, Identifier, Identifier] | Tuple[Identifier, identifier] | Node
    ]: ...
    def __contains__(self, triple: Any) -> bool: ...
    def __add__(self, other: Any) -> Graph: ...
    def set(self, triple: Any) -> None: ...
    def subjects(
        self, predicate: Optional[Any] = ..., object: Optional[Any] = ...
    ) -> Iterable[Node]: ...
    def predicates(
        self, subject: Optional[Any] = ..., object: Optional[Any] = ...
    ) -> Iterable[Node]: ...
    def objects(
        self, subject: Optional[Any] = ..., predicate: Optional[Any] = ...
    ) -> Iterable[Identifier]: ...
    def subject_predicates(self, object: Optional[Any] = ...) -> None: ...
    def subject_objects(self, predicate: Optional[Any] = ...) -> None: ...
    def predicate_objects(self, subject: Optional[Any] = ...) -> None: ...
    def triples_choices(self, triple: Any, context: Optional[Any] = ...) -> None: ...
    def value(
        self,
        subject: Optional[Any] = ...,
        predicate: Any = ...,
        object: Optional[Any] = ...,
        default: Optional[Any] = ...,
        any: bool = ...,
    ) -> Any: ...
    def label(self, subject: Any, default: str = ...) -> Any: ...
    def preferredLabel(
        self,
        subject: Any,
        lang: Optional[Any] = ...,
        default: Optional[Any] = ...,
        labelProperties: Any = ...,
    ) -> List[Tuple[Any, Any]]: ...
    def comment(self, subject: Any, default: str = ...) -> Any: ...
    def items(self, list: Any) -> Iterator[Any]: ...
    def transitiveClosure(
        self, func: Any, arg: Any, seen: Optional[Any] = ...
    ) -> Iterator[Any]: ...
    def transitive_objects(
        self, subject: Any, property: Any, remember: Optional[Any] = ...
    ) -> Iterator[Any]: ...
    def transitive_subjects(
        self, predicate: Any, object: Any, remember: Optional[Any] = ...
    ) -> Iterator[Any]: ...
    def seq(self, subject: Any) -> Seq | None: ...
    def qname(self, uri: Any) -> Any: ...
    def compute_qname(self, uri: Any, generate: bool = ...) -> Any: ...
    def bind(
        self, prefix: Any, namespace: Any, override: bool = ..., replace: bool = ...
    ) -> Any: ...
    def namespaces(self) -> Iterator[Tuple[Any, Any]]: ...
    def absolutize(self, uri: Any, defrag: int = ...) -> Any: ...

    # no destination and non-None positional encoding
    @overload
    def serialize(
        self,
        destination: None,
        format: str,
        base: Optional[str],
        encoding: str,
        **args: Any,
    ) -> bytes: ...

    # no destination and non-None keyword encoding
    @overload
    def serialize(
        self,
        destination: None = ...,
        format: str = ...,
        base: Optional[str] = ...,
        *,
        encoding: str,
        **args: Any,
    ) -> bytes: ...

    # no destination and None encoding
    @overload
    def serialize(
        self,
        destination: None = ...,
        format: str = ...,
        base: Optional[str] = ...,
        encoding: None = ...,
        **args: Any,
    ) -> str: ...

    # non-None destination
    @overload
    def serialize(
        self,
        destination: Union[str, pathlib.PurePath, IO[bytes]],
        format: str = ...,
        base: Optional[str] = ...,
        encoding: Optional[str] = ...,
        **args: Any,
    ) -> "Graph": ...

    # fallback
    @overload
    def serialize(
        self,
        destination: Optional[Union[str, pathlib.PurePath, IO[bytes]]] = ...,
        format: str = ...,
        base: Optional[str] = ...,
        encoding: Optional[str] = ...,
        **args: Any,
    ) -> Union[bytes, str, "Graph"]: ...
    def parse(
        self,
        source: Optional[Any] = ...,
        publicID: Optional[Any] = ...,
        format: Optional[str] = ...,
        location: Optional[Any] = ...,
        file: Optional[Any] = ...,
        data: Optional[Any] = ...,
        **args: Any,
    ) -> "Graph": ...
    def load(self, source: Any, publicID: Optional[Any] = ..., format: str = ...) -> "Graph": ...
    def query(
        self,
        query_object: Any,
        processor: str = ...,
        result: str = ...,
        initNs: Optional[Any] = ...,
        initBindings: Optional[Any] = ...,
        use_store_provided: bool = ...,
        **kwargs: Any,
    ) -> query.Result: ...
    def update(
        self,
        update_object: Any,
        processor: str = ...,
        initNs: Optional[Any] = ...,
        initBindings: Optional[Any] = ...,
        use_store_provided: bool = ...,
        **kwargs: Any,
    ) -> Any: ...
    def n3(self) -> str: ...
    def isomorphic(self, other: Any) -> bool: ...
    def connected(self) -> bool: ...
    def all_nodes(self) -> Set[Any]: ...
    def collection(self, identifier: Any) -> Collection: ...
    def resource(self, identifier: Any) -> Resource: ...
    def skolemize(
        self,
        new_graph: Optional[Any] = ...,
        bnode: Optional[Any] = ...,
        authority: Optional[Any] = ...,
        basepath: Optional[Any] = ...,
    ) -> Graph: ...
    def de_skolemize(
        self, new_graph: Optional[Any] = ..., uriref: Optional[Any] = ...
    ) -> Graph: ...

class ConjunctiveGraph(Graph):
    context_aware: bool = ...
    default_union: bool = ...
    default_context: Any = ...
    def __init__(
        self,
        store: str = ...,
        identifier: Optional[Any] = ...,
        default_graph_base: Optional[Any] = ...,
    ) -> None: ...
    def add(self, triple_or_quad: Any) -> None: ...
    def addN(self, quads: Any) -> None: ...
    def remove(self, triple_or_quad: Any) -> None: ...
    # def triples(self, triple_or_quad: Tuple[Optional[Union[str, BNode]], Optional[Union[str, BNode]], Optional[BNode]], context: Tuple[Optional[Union[str, BNode]], Optional[Union[str, BNode]], Optional[BNode]]) -> Iterator[Tuple[Identifier, Identifier, Identifier]]: ...
    def quads(self, triple_or_quad: Optional[Any] = ...) -> None: ...
    def triples_choices(self, triple: Any, context: Optional[Any] = ...) -> None: ...
    def contexts(self, triple: Optional[Any] = ...) -> None: ...
    def get_context(
        self,
        identifier: Node | str | None,
        quoted: bool = ...,
        base: Optional[str] = ...,
    ) -> Graph: ...
    def remove_context(self, context: Any) -> None: ...
    def context_id(self, uri: Any, context_id: Optional[Any] = ...) -> Any: ...
    def parse(
        self,
        source: Optional[Any] = ...,
        publicID: Optional[Any] = ...,
        format: Optional[str] = ...,
        location: Optional[Any] = ...,
        file: Optional[Any] = ...,
        data: Optional[Any] = ...,
        **args: Any,
    ) -> Graph: ...

class Seq:
    def __init__(self, graph: Graph, subject: Any) -> None: ...
    def toPython(self) -> Seq: ...



from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class QB(DefinedNamespace):
    attribute: URIRef
    codeList: URIRef
    component: URIRef
    componentAttachment: URIRef
    componentProperty: URIRef
    componentRequired: URIRef
    concept: URIRef
    dataSet: URIRef
    dimension: URIRef
    hierarchyRoot: URIRef
    measure: URIRef
    measureDimension: URIRef
    measureType: URIRef
    observation: URIRef
    observationGroup: URIRef
    order: URIRef
    parentChildProperty: URIRef
    slice: URIRef
    sliceKey: URIRef
    sliceStructure: URIRef
    structure: URIRef
    Attachable: URIRef
    AttributeProperty: URIRef
    CodedProperty: URIRef
    ComponentProperty: URIRef
    ComponentSet: URIRef
    ComponentSpecification: URIRef
    DataSet: URIRef
    DataStructureDefinition: URIRef
    DimensionProperty: URIRef
    HierarchicalCodeList: URIRef
    MeasureProperty: URIRef
    Observation: URIRef
    ObservationGroup: URIRef
    Slice: URIRef
    SliceKey: URIRef



from typing import Any, Callable, ClassVar, Generic, List, Optional, Set, TypeVar, Union

from typing_extensions import Literal, NotRequired, Required, TypeAlias, TypedDict

AstType = Literal[
    "text",
    "image",
    "codespan",
    "linebreak",
    "inline_html",
    "heading",
    "newline",
    "thematic_break",
    "block_code",
    "block_html",
    "list",
    "list_item",
]
_AstChildren: TypeAlias = "AstChildren"
AstObject = TypedDict(
    "AstObject",
    {
        "type": Required[AstType],
        "text": Required[Optional[_AstChildren]],
        "link": NotRequired[str],
        "info": NotRequired[Optional[str]],
        "children": NotRequired[Union[_AstChildren, List[_AstChildren]]],
        "ordered": NotRequired[bool],
        "level": NotRequired[int],
    },
    total=False,
)
AstChildren = Union[AstObject, str]

HTMLType = str
DataT = TypeVar("DataT", bound=Union[AstChildren, HTMLType])

RenderMethod = Union[
    Callable[[], DataT],  # blank
    Callable[[DataT, Any], DataT],
]

class BaseRenderer(Generic[DataT]):
    NAME: ClassVar[str] = "base"

    def _get_method(self, name: AstType) -> RenderMethod[DataT]: ...
    def finalize(self, data: DataT) -> Any: ...

class AstRenderer(BaseRenderer[AstChildren]):
    NAME: ClassVar[Literal["ast"]] = "ast"

    def __init__(
        self, escape: bool = True, allow_harmful_protocols: Optional[bool] = None
    ) -> None: ...
    def _create_default_method(self, name: AstType) -> RenderMethod[AstChildren]: ...
    def _get_method(self, name: AstType) -> RenderMethod[AstChildren]: ...
    def text(self, text: str) -> AstChildren: ...
    def link(
        self, link: str, text: Optional[str] = None, title: Optional[str] = None
    ) -> AstChildren: ...
    def image(self, src: str, alt: str = "", title: Optional[str] = None) -> AstChildren: ...
    def codespan(self, text: str) -> AstChildren: ...
    def linebreak(self, text: str) -> AstChildren: ...
    def inline_html(self, text: str) -> AstChildren: ...
    def paragraph(self, text: str) -> AstChildren: ...
    def heading(self, text: str, level: int) -> AstChildren: ...
    def newline(self) -> AstChildren: ...
    def thematic_break(self) -> AstChildren: ...
    def block_code(self, code: str, info: Optional[str] = None) -> AstChildren: ...
    def block_html(self, html: str) -> AstChildren: ...
    def list(
        self, children: str, ordered: bool, level: int, start: Optional[int] = None
    ) -> AstChildren: ...
    def list_item(self, children: str, level: int) -> AstChildren: ...
    def finalize(self, data: DataT) -> List[DataT]: ...

class HTMLRenderer(BaseRenderer[HTMLType]):
    NAME: ClassVar[Literal["html"]] = "html"
    HARMFUL_PROTOCOLS: ClassVar[Set[str]]
    _escape: bool
    _allow_harmful_protocols: Optional[bool]

    def __init__(
        self, escape: bool = True, allow_harmful_protocols: Optional[bool] = None
    ) -> None: ...
    def _safe_url(self, url: str) -> str: ...
    def text(self, text: str) -> HTMLType: ...
    def link(
        self, link: str, text: Optional[str] = None, title: Optional[str] = None
    ) -> HTMLType: ...
    def image(self, src: str, alt: str = "", title: Optional[str] = None) -> HTMLType: ...
    def emphasis(self, text: str) -> HTMLType: ...
    def strong(self, text: str) -> HTMLType: ...
    def codespan(self, text: str) -> HTMLType: ...
    def linebreak(self, text: str) -> HTMLType: ...
    def inline_html(self, text: str) -> HTMLType: ...
    def paragraph(self, text: str) -> HTMLType: ...
    def heading(self, text: str, level: int) -> HTMLType: ...
    def newline(self) -> HTMLType: ...
    def thematic_break(self) -> HTMLType: ...
    def block_code(self, code: str, info: Optional[str] = None) -> HTMLType: ...
    def block_quote(self, text: str) -> HTMLType: ...
    def block_html(self, html: str) -> HTMLType: ...
    def block_error(self, html: str) -> HTMLType: ...
    def list(
        self, text: str, ordered: bool, level: int, start: Optional[int] = None
    ) -> HTMLType: ...
    def list_item(self, text: str, level: int) -> HTMLType: ...
    def finalize(self, data: DataT) -> HTMLType: ...



from typing import Collection

from _typeshed import Incomplete

from .cache import BaseCache

class CacheController:
    cache: BaseCache
    cache_etags: bool
    serializer: Incomplete
    cacheable_status_codes: Collection[int] | None = None
    def __init__(
        self,
        cache: BaseCache | None = None,
        cache_etags: bool = True,
        serializer: Incomplete | None = None,
        status_codes: Collection[int] | None = None,
    ) -> None: ...



"""
C++17 code generator for a given Schema Salad definition.

Currently only supports emiting YAML from the C++ objects, not yet parsing
YAML into C++ objects.

The generated code requires the libyaml-cpp library & headers

To see an example of usage, look at schema_salad/tests/codegen/cwl.cpp
which can be combined with the CWL V1.0 schema as shown below::

  schema-salad-tool --codegen cpp \
          schema_salad/tests/test_schema/CommonWorkflowLanguage.yml \
          > cwl_v1_0.h

  g++ --std=c++20 -I. -lyaml-cpp schema_salad/tests/codegen/cwl.cpp -o cwl-v1_0-test
  ./cwl-v1_0-test

  # g++ versions older than version 10 may need "--std=c++2a" instead of "--std=c++20"
"""
import re
from typing import IO, Any, Dict, List, Optional, Tuple, Union, cast

from . import _logger
from .codegen_base import CodeGenBase, TypeDef
from .exceptions import SchemaException
from .schema import shortname
from .utils import aslist


def replaceKeywords(s: str) -> str:
    """Rename keywords that are reserved in C++."""
    if s in (
        "class",
        "enum",
        "int",
        "long",
        "float",
        "double",
        "default",
        "stdin",
        "stdout",
        "stderr",
    ):
        s = s + "_"
    return s


def safename(name: str) -> str:
    """Create a C++ safe name."""
    classname = re.sub("[^a-zA-Z0-9]", "_", name)
    return replaceKeywords(classname)


# TODO: this should be somehow not really exists
def safename2(name: Dict[str, str]) -> str:
    """Create a namespaced safename."""
    return safename(name["namespace"]) + "::" + safename(name["classname"])


# Splits names like https://xyz.xyz/blub#cwl/class
# into its class path and non class path
def split_name(s: str) -> Tuple[str, str]:
    t = s.split("#")
    if len(t) != 2:
        raise ValueError("Expected field to be formatted as 'https://xyz.xyz/blub#cwl/class'.")
    return (t[0], t[1])


# similar to split_name but for field names
def split_field(s: str) -> Tuple[str, str, str]:
    (namespace, field) = split_name(s)
    t = field.split("/")
    if len(t) != 2:
        raise ValueError("Expected field to be formatted as 'https://xyz.xyz/blub#cwl/class'.")
    return (namespace, t[0], t[1])


# Prototype of a class
class ClassDefinition:
    def __init__(self, name: str):
        self.fullName = name
        self.extends: List[Dict[str, str]] = []
        self.fields: List[FieldDefinition] = []
        self.abstract = False
        (self.namespace, self.classname) = split_name(name)
        self.namespace = safename(self.namespace)
        self.classname = safename(self.classname)

    def writeFwdDeclaration(self, target: IO[str], fullInd: str, ind: str) -> None:
        target.write(f"{fullInd}namespace {self.namespace} {{ struct {self.classname}; }}\n")

    def writeDefinition(self, target: IO[Any], fullInd: str, ind: str) -> None:
        target.write(f"{fullInd}namespace {self.namespace} {{\n")
        target.write(f"{fullInd}struct {self.classname}")
        extends = list(map(safename2, self.extends))
        override = ""
        virtual = "virtual "
        if len(self.extends) > 0:
            target.write(f"\n{fullInd}{ind}: ")
            target.write(f"\n{fullInd}{ind}, ".join(extends))
            override = " override"
            virtual = ""
        target.write(" {\n")

        for field in self.fields:
            field.writeDefinition(target, fullInd + ind, ind, self.namespace)

        if self.abstract:
            target.write(f"{fullInd}{ind}virtual ~{self.classname}() = 0;\n")
        target.write(f"{fullInd}{ind}{virtual}auto toYaml() const -> YAML::Node{override};\n")
        target.write(f"{fullInd}}};\n")
        target.write(f"{fullInd}}}\n\n")

    def writeImplDefinition(self, target: IO[str], fullInd: str, ind: str) -> None:
        extends = list(map(safename2, self.extends))

        if self.abstract:
            target.write(
                f"{fullInd}inline {self.namespace}::{self.classname}::~{self.classname}() = default;\n"
            )

        target.write(
            f"""{fullInd}inline auto {self.namespace}::{self.classname}::toYaml() const -> YAML::Node {{
{fullInd}{ind}using ::toYaml;
{fullInd}{ind}auto n = YAML::Node{{}};
"""
        )
        for e in extends:
            target.write(f"{fullInd}{ind}n = mergeYaml(n, {e}::toYaml());\n")

        for field in self.fields:
            fieldname = safename(field.name)
            target.write(
                f'{fullInd}{ind}addYamlField(n, "{field.name}", toYaml(*{fieldname}));\n'  # noqa: B907
            )
            # target.write(f"{fullInd}{ind}addYamlIfNotEmpty(n, \"{field.name}\", toYaml(*{fieldname}));\n")

        target.write(f"{fullInd}{ind}return n;\n{fullInd}}}\n")


# Prototype of a single field of a class
class FieldDefinition:
    def __init__(self, name: str, typeStr: str, optional: bool):
        self.name = name
        self.typeStr = typeStr
        self.optional = optional

    def writeDefinition(self, target: IO[Any], fullInd: str, ind: str, namespace: str) -> None:
        """Write a C++ definition for the class field."""
        name = safename(self.name)
        typeStr = self.typeStr.replace(namespace + "::", "")
        target.write(f"{fullInd}heap_object<{typeStr}> {name};\n")


# Prototype of an enum definition
class EnumDefinition:
    def __init__(self, name: str, values: List[str]):
        self.name = name
        self.values = values

    def writeDefinition(self, target: IO[str], ind: str) -> None:
        namespace = ""
        if len(self.name.split("#")) == 2:
            (namespace, classname) = split_name(self.name)
            namespace = safename(namespace)
            classname = safename(classname)
            name = namespace + "::" + classname
        else:
            name = safename(self.name)
            classname = name
        if len(namespace) > 0:
            target.write(f"namespace {namespace} {{\n")
        target.write(f"enum class {classname} : unsigned int {{\n{ind}")
        target.write(f",\n{ind}".join(map(safename, self.values)))
        target.write("\n};\n")
        target.write(f"inline auto to_string({classname} v) {{\n")
        target.write(f"{ind}static auto m = std::vector<std::string_view> {{\n")
        target.write(f'{ind}    "')
        target.write(f'",\n{ind}    "'.join(self.values))
        target.write(f'"\n{ind}}};\n')

        target.write(f"{ind}using U = std::underlying_type_t<{name}>;\n")
        target.write(f"{ind}return m.at(static_cast<U>(v));\n}}\n")

        if len(namespace) > 0:
            target.write("}\n")

        target.write(f"inline void to_enum(std::string_view v, {name}& out) {{\n")
        target.write(f"{ind}static auto m = std::map<std::string, {name}, std::less<>> {{\n")
        for v in self.values:
            target.write(f'{ind}{ind}{{"{v}", {name}::{safename(v)}}},\n')  # noqa: B907
        target.write(f"{ind}}};\n{ind}out = m.find(v)->second;\n}}\n")

        target.write(f"inline auto toYaml({name} v) {{\n")
        target.write(f"{ind}return YAML::Node{{std::string{{to_string(v)}}}};\n}}\n")

        target.write(f"inline auto yamlToEnum(YAML::Node n, {name}& out) {{\n")
        target.write(f"{ind}to_enum(n.as<std::string>(), out);\n}}\n")


# !TODO way tot many functions, most of these shouldn't exists
def isPrimitiveType(v: Any) -> bool:
    if not isinstance(v, str):
        return False
    return v in ["null", "boolean", "int", "long", "float", "double", "string"]


def hasFieldValue(e: Any, f: str, v: Any) -> bool:
    if not isinstance(e, dict):
        return False
    if f not in e:
        return False
    return bool(e[f] == v)


def isRecordSchema(v: Any) -> bool:
    return hasFieldValue(v, "type", "record")


def isEnumSchema(v: Any) -> bool:
    if not hasFieldValue(v, "type", "enum"):
        return False
    if "symbols" not in v:
        return False
    if not isinstance(v["symbols"], list):
        return False
    return True


def isArray(v: Any) -> bool:
    if not isinstance(v, list):
        return False
    for i in v:
        if not pred(i):
            return False
    return True


def pred(i: Any) -> bool:
    return (
        isPrimitiveType(i)
        or isRecordSchema(i)
        or isEnumSchema(i)
        or isArraySchema(i)
        or isinstance(i, str)
    )


def isArraySchema(v: Any) -> bool:
    if not hasFieldValue(v, "type", "array"):
        return False
    if "items" not in v:
        return False
    if not isinstance(v["items"], list):
        return False

    for i in v["items"]:
        if not (pred(i) or isArray(i)):
            return False
    return True


class CppCodeGen(CodeGenBase):
    """Generation of C++ code for a given Schema Salad definition."""

    def __init__(
        self,
        base: str,
        target: IO[str],
        examples: Optional[str],
        package: str,
        copyright: Optional[str],
    ) -> None:
        super().__init__()
        self.base_uri = base
        self.target = target
        self.examples = examples
        self.package = package
        self.copyright = copyright

        self.classDefinitions: Dict[str, ClassDefinition] = {}
        self.enumDefinitions: Dict[str, EnumDefinition] = {}

    def convertTypeToCpp(self, type_declaration: Union[List[Any], Dict[str, Any], str]) -> str:
        """Convert a Schema Salad type to a C++ type."""
        if not isinstance(type_declaration, list):
            return self.convertTypeToCpp([type_declaration])

        if len(type_declaration) == 1:
            if type_declaration[0] in ("null", "https://w3id.org/cwl/salad#null"):
                return "std::monostate"
            elif type_declaration[0] in (
                "string",
                "http://www.w3.org/2001/XMLSchema#string",
            ):
                return "std::string"
            elif type_declaration[0] in ("int", "http://www.w3.org/2001/XMLSchema#int"):
                return "int32_t"
            elif type_declaration[0] in (
                "long",
                "http://www.w3.org/2001/XMLSchema#long",
            ):
                return "int64_t"
            elif type_declaration[0] in (
                "float",
                "http://www.w3.org/2001/XMLSchema#float",
            ):
                return "float"
            elif type_declaration[0] in (
                "double",
                "http://www.w3.org/2001/XMLSchema#double",
            ):
                return "double"
            elif type_declaration[0] in (
                "boolean",
                "http://www.w3.org/2001/XMLSchema#boolean",
            ):
                return "bool"
            elif type_declaration[0] == "https://w3id.org/cwl/salad#Any":
                return "std::any"
            elif type_declaration[0] in (
                "PrimitiveType",
                "https://w3id.org/cwl/salad#PrimitiveType",
            ):
                return "std::variant<bool, int32_t, int64_t, float, double, std::string>"
            elif isinstance(type_declaration[0], dict):
                if "type" in type_declaration[0] and type_declaration[0]["type"] in (
                    "enum",
                    "https://w3id.org/cwl/salad#enum",
                ):
                    name = type_declaration[0]["name"]
                    if name not in self.enumDefinitions:
                        self.enumDefinitions[name] = EnumDefinition(
                            type_declaration[0]["name"],
                            list(map(shortname, type_declaration[0]["symbols"])),
                        )
                    if len(name.split("#")) != 2:
                        return safename(name)
                    (namespace, classname) = name.split("#")
                    return safename(namespace) + "::" + safename(classname)
                elif "type" in type_declaration[0] and type_declaration[0]["type"] in (
                    "array",
                    "https://w3id.org/cwl/salad#array",
                ):
                    items = type_declaration[0]["items"]
                    if isinstance(items, list):
                        ts = []
                        for i in items:
                            ts.append(self.convertTypeToCpp(i))
                        name = ", ".join(ts)
                        return f"std::vector<std::variant<{name}>>"
                    else:
                        i = self.convertTypeToCpp(items)
                        return f"std::vector<{i}>"
                elif "type" in type_declaration[0] and type_declaration[0]["type"] in (
                    "record",
                    "https://w3id.org/cwl/salad#record",
                ):
                    n = type_declaration[0]["name"]
                    (namespace, classname) = split_name(n)
                    return safename(namespace) + "::" + safename(classname)

                n = type_declaration[0]["type"]
                (namespace, classname) = split_name(n)
                return safename(namespace) + "::" + safename(classname)

            if len(type_declaration[0].split("#")) != 2:
                _logger.debug(f"// something weird2 about {type_declaration[0]}")
                return cast(str, type_declaration[0])

            (namespace, classname) = split_name(type_declaration[0])
            return safename(namespace) + "::" + safename(classname)

        type_declaration = list(map(self.convertTypeToCpp, type_declaration))
        type_declaration = ", ".join(type_declaration)
        return f"std::variant<{type_declaration}>"

    # start of our generated file
    def epilogue(self, root_loader: Optional[TypeDef]) -> None:
        self.target.write(
            """#pragma once

// Generated by schema-salad code generator

#include <cassert>
#include <cstddef>
#include <cstdint>
#include <map>
#include <string>
#include <string_view>
#include <variant>
#include <vector>
#include <yaml-cpp/yaml.h>
#include <any>

inline auto mergeYaml(YAML::Node n1, YAML::Node n2) {
    for (auto const& e : n1) {
        n2[e.first.as<std::string>()] = e.second;
    }
    return n2;
}

// declaring toYaml
inline auto toYaml(bool v) {
    return YAML::Node{v};
}
inline auto toYaml(float v) {
    return YAML::Node{v};
}
inline auto toYaml(double v) {
    return YAML::Node{v};
}
inline auto toYaml(int32_t v) {
    return YAML::Node{v};
}
inline auto toYaml(int64_t v) {
    return YAML::Node{v};
}
inline auto toYaml(std::any const&) {
    return YAML::Node{};
}
inline auto toYaml(std::monostate const&) {
    return YAML::Node(YAML::NodeType::Undefined);
}

inline auto toYaml(std::string const& v) {
    return YAML::Node{v};
}

inline void addYamlField(YAML::Node node, std::string const& key, YAML::Node value) {
    if (value.IsDefined()) {
        node[key] = value;
    }
}

// fwd declaring toYaml
template <typename T>
auto toYaml(std::vector<T> const& v) -> YAML::Node;
template <typename T>
auto toYaml(T const& t) -> YAML::Node;
template <typename ...Args>
auto toYaml(std::variant<Args...> const& t) -> YAML::Node;

template <typename T>
class heap_object {
    std::unique_ptr<T> data = std::make_unique<T>();
public:
    heap_object() = default;
    heap_object(heap_object const& oth) {
        *data = *oth;
    }
    heap_object(heap_object&& oth) {
        *data = *oth;
    }

    template <typename T2>
    heap_object(T2 const& oth) {
        *data = oth;
    }
    template <typename T2>
    heap_object(T2&& oth) {
        *data = oth;
    }

    auto operator=(heap_object const& oth) -> heap_object& {
        *data = *oth;
        return *this;
    }
    auto operator=(heap_object&& oth) -> heap_object& {
        *data = std::move(*oth);
        return *this;
    }

    template <typename T2>
    auto operator=(T2 const& oth) -> heap_object& {
        *data = oth;
        return *this;
    }
    template <typename T2>
    auto operator=(T2&& oth) -> heap_object& {
        *data = std::move(oth);
        return *this;
    }

    auto operator->() -> T* {
        return data.get();
    }
    auto operator->() const -> T const* {
        return data.get();
    }
    auto operator*() -> T& {
        return *data;
    }
    auto operator*() const -> T const& {
        return *data;
    }

};

"""
        )
        # main body, printing fwd declaration, class definitions, and then implementations

        for key in self.classDefinitions:
            self.classDefinitions[key].writeFwdDeclaration(self.target, "", "    ")

        for key in self.enumDefinitions:
            self.enumDefinitions[key].writeDefinition(self.target, "    ")
        for key in self.classDefinitions:
            self.classDefinitions[key].writeDefinition(self.target, "", "    ")
        for key in self.classDefinitions:
            self.classDefinitions[key].writeImplDefinition(self.target, "", "    ")

        self.target.write(
            """
template <typename T>
auto toYaml(std::vector<T> const& v) -> YAML::Node {
    auto n = YAML::Node(YAML::NodeType::Sequence);
    for (auto const& e : v) {
        n.push_back(toYaml(e));
    }
    return n;
}

template <typename T>
auto toYaml(T const& t) -> YAML::Node {
    if constexpr (std::is_enum_v<T>) {
        return toYaml(t);
    } else {
        return t.toYaml();
    }
}

template <typename ...Args>
auto toYaml(std::variant<Args...> const& t) -> YAML::Node {
    return std::visit([](auto const& e) {
        return toYaml(e);
    }, t);
}
"""
        )

    def parseRecordField(self, field: Dict[str, Any]) -> FieldDefinition:
        (namespace, classname, fieldname) = split_field(field["name"])
        if isinstance(field["type"], dict):
            if field["type"]["type"] == "enum":
                fieldtype = "Enum"
            else:
                fieldtype = self.convertTypeToCpp(field["type"])

        else:
            fieldtype = field["type"]
            fieldtype = self.convertTypeToCpp(fieldtype)

        return FieldDefinition(name=fieldname, typeStr=fieldtype, optional=False)

    def parseRecordSchema(self, stype: Dict[str, Any]) -> None:
        cd = ClassDefinition(name=stype["name"])
        cd.abstract = stype.get("abstract", False)

        if "extends" in stype:
            for ex in aslist(stype["extends"]):
                (base_namespace, base_classname) = split_name(ex)
                ext = {"namespace": base_namespace, "classname": base_classname}
                cd.extends.append(ext)

        if "fields" in stype:
            for field in stype["fields"]:
                cd.fields.append(self.parseRecordField(field))

        self.classDefinitions[stype["name"]] = cd

    def parseEnum(self, stype: Dict[str, Any]) -> str:
        name = cast(str, stype["name"])
        if name not in self.enumDefinitions:
            self.enumDefinitions[name] = EnumDefinition(
                name, list(map(shortname, stype["symbols"]))
            )
        return name

    def parse(self, items: List[Dict[str, Any]]) -> None:
        for stype in items:
            if "type" in stype and stype["type"] == "documentation":
                continue

            if not (pred(stype) or isArray(stype)):
                raise SchemaException("not a valid SaladRecordField")

            # parsing a record
            if isRecordSchema(stype):
                self.parseRecordSchema(stype)
            elif isEnumSchema(stype):
                self.parseEnum(stype)
            else:
                _logger.error(f"not parsed{stype}")

        self.epilogue(None)
        self.target.close()



"""Test examples."""
import datetime
import os
from io import StringIO
from typing import Any, Dict, cast

import pytest
from ruamel.yaml.comments import CommentedMap, CommentedSeq

import schema_salad.main
import schema_salad.schema
from schema_salad.jsonld_context import makerdf
from schema_salad.ref_resolver import Loader, file_uri, uri_file_path
from schema_salad.sourceline import SourceLine, cmap
from schema_salad.utils import ContextType, stdout, yaml_no_ts

from .util import get_data


def test_schemas() -> None:
    loader = Loader({})

    path = get_data("tests/EDAM.owl")
    assert path
    ra, _ = loader.resolve_all(
        cmap(
            {
                "$schemas": [file_uri(path)],
                "$namespaces": {"edam": "http://edamontology.org/"},
                "edam:has_format": "edam:format_1915",
            }
        ),
        "",
    )

    assert {
        "$schemas": [file_uri(path)],
        "$namespaces": {"edam": "http://edamontology.org/"},
        "http://edamontology.org/has_format": "http://edamontology.org/format_1915",
    } == ra


def test_bad_schemas(caplog: pytest.LogCaptureFixture) -> None:
    """Test that bad $schemas refs don't stop parsing."""
    schema_path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_path = get_data("tests/revtool_bad_schema.cwl")
    assert schema_path and document_path
    assert 0 == schema_salad.main.main(
        argsl=["--print-rdf", schema_path, document_path],
    )
    assert (
        "Could not load extension schema https://bad.example.com/missing.ttl: "
        "Error fetching https://bad.example.com/missing.ttl"
    ) in caplog.text
    assert (
        "https://schema.org/!DOCTYPE html does not look like a valid URI, "
        "trying to serialize this will break"
    ) in caplog.text
    assert (
        "Could not load extension schema https://schema.org/docs/schema_org_rdfa.html: "
        "No plugin registered for (rdfa, <class 'rdflib.parser.Parser'>)"
    ) in caplog.text


def test_skip_bad_schemas(caplog: pytest.LogCaptureFixture) -> None:
    """Test that (bad) $schemas refs are properly skipped."""
    schema_path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_path = get_data("tests/revtool_bad_schema.cwl")
    assert schema_path and document_path
    assert 0 == schema_salad.main.main(
        argsl=["--print-rdf", "--skip-schemas", schema_path, document_path],
    )
    assert (
        "Could not load extension schema https://bad.example.com/missing.ttl: "
        "Error fetching https://bad.example.com/missing.ttl"
    ) not in caplog.text
    assert (
        "https://schema.org/!DOCTYPE html does not look like a valid URI, "
        "trying to serialize this will break"
    ) not in caplog.text
    assert (
        "Could not load extension schema https://schema.org/docs/schema_org_rdfa.html: "
        "No plugin registered for (rdfa, <class 'rdflib.parser.Parser'>)"
    ) not in caplog.text
    assert "rdflib.plugin.PluginException: No plugin registered" not in caplog.text


def test_self_validate() -> None:
    path = get_data("metaschema/metaschema.yml")
    assert path
    assert 0 == schema_salad.main.main(argsl=[path])
    assert 0 == schema_salad.main.main(argsl=[path, path])


def test_print_rdf() -> None:
    """Test --print-rdf."""
    schema_path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_path = get_data("tests/test_real_cwl/bio-cwl-tools/bamtools_stats.cwl")
    assert schema_path and document_path
    assert 0 == schema_salad.main.main(argsl=["--print-rdf", schema_path, document_path])


def test_print_rdf_invalid_external_ref() -> None:
    """Test --print-rdf when document references unfetchable external schema."""
    schema_path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_path = get_data(
        "tests/test_real_cwl/bio-cwl-tools/bamtools_stats_invalid_schema_ref.cwl"
    )
    assert schema_path and document_path
    assert 0 == schema_salad.main.main(argsl=["--print-rdf", schema_path, document_path])


def test_print_pre_schema() -> None:
    """Test --print-pre only schema."""
    schema_path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    assert schema_path
    assert 0 == schema_salad.main.main(argsl=["--print-pre", schema_path])


def test_print_pre() -> None:
    """Test --print-pre."""
    schema_path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_path = get_data("tests/test_real_cwl/bio-cwl-tools/bamtools_stats.cwl")
    assert schema_path and document_path
    assert 0 == schema_salad.main.main(argsl=["--print-pre", schema_path, document_path])


def test_print_schema_index() -> None:
    """Test --print-index only with a schema."""
    schema_path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    assert schema_path
    assert 0 == schema_salad.main.main(argsl=["--print-index", schema_path])


def test_print_index() -> None:
    """Test --print-index."""
    schema_path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_path = get_data("tests/test_real_cwl/bio-cwl-tools/bamtools_stats.cwl")
    assert schema_path and document_path
    assert 0 == schema_salad.main.main(argsl=["--print-index", schema_path, document_path])


def test_print_schema_metadata() -> None:
    """Test --print-metadata only for a schema."""
    schema_path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    assert schema_path
    assert 0 == schema_salad.main.main(argsl=["--print-metadata", schema_path])


def test_print_metadata() -> None:
    """Test --print-metadata."""
    schema_path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
    document_path = get_data("tests/test_real_cwl/bio-cwl-tools/bamtools_stats.cwl")
    assert schema_path and document_path
    assert 0 == schema_salad.main.main(argsl=["--print-metadata", schema_path, document_path])


def test_schema_salad_doc_oneline_doc() -> None:
    """Test schema-salad-doc when the 1st type has only a single doc line."""
    schema_path = get_data("tests/test_schema/one_line_primary_doc.yml")
    assert schema_path
    stdout = StringIO()
    schema_salad.makedoc.makedoc(stdout, schema_path)
    assert "<title>Schema</title>" in stdout.getvalue()


def test_avro_regression() -> None:
    path = get_data("tests/Process.yml")
    assert path
    assert 0 == schema_salad.main.main(argsl=[path])


def test_jsonld_ctx() -> None:
    ldr, _, _, _ = schema_salad.schema.load_schema(
        cmap(
            {
                "$base": "Y",
                "name": "X",
                "$namespaces": {"foo": "http://example.com/foo#"},
                "$graph": [{"name": "ExampleType", "type": "enum", "symbols": ["asym", "bsym"]}],
            }
        )
    )

    ra, _ = ldr.resolve_all(cmap({"foo:bar": "asym"}), "X")

    assert ra == {"http://example.com/foo#bar": "asym"}


def test_idmap() -> None:
    ldr = Loader({})
    ldr.add_context(
        {
            "inputs": {
                "@id": "http://example.com/inputs",
                "mapSubject": "id",
                "mapPredicate": "a",
            },
            "outputs": {"@type": "@id", "identity": True},
            "id": "@id",
        }
    )

    ra, _ = ldr.resolve_all(
        cmap(
            {
                "id": "stuff",
                "inputs": {"zip": 1, "zing": 2},
                "outputs": ["out"],
                "other": {"n": 9},
            }
        ),
        "http://example2.com/",
    )
    assert isinstance(ra, CommentedMap)

    assert "http://example2.com/#stuff" == ra["id"]
    for item in ra["inputs"]:
        if item["a"] == 2:
            assert "http://example2.com/#stuff/zing" == item["id"]
        else:
            assert "http://example2.com/#stuff/zip" == item["id"]
    assert ["http://example2.com/#stuff/out"] == ra["outputs"]
    assert {"n": 9} == ra["other"]


def test_scoped_ref() -> None:
    ldr = Loader({})
    ldr.add_context(
        {
            "scatter": {"@type": "@id", "refScope": 0},
            "source": {"@type": "@id", "refScope": 2},
            "in": {"mapSubject": "id", "mapPredicate": "source"},
            "out": {"@type": "@id", "identity": True},
            "inputs": {"mapSubject": "id", "mapPredicate": "type"},
            "outputs": {"mapSubject": "id"},
            "steps": {"mapSubject": "id"},
            "id": "@id",
        }
    )

    ra, _ = ldr.resolve_all(
        cmap(
            {
                "inputs": {"inp": "string", "inp2": "string"},
                "outputs": {"out": {"type": "string", "source": "step2/out"}},
                "steps": {
                    "step1": {
                        "in": {"inp": "inp", "inp2": "#inp2", "inp3": ["inp", "inp2"]},
                        "out": ["out"],
                        "scatter": "inp",
                    },
                    "step2": {
                        "in": {"inp": "step1/out"},
                        "scatter": "inp",
                        "out": ["out"],
                    },
                },
            }
        ),
        "http://example2.com/",
    )

    assert {
        "inputs": [
            {"id": "http://example2.com/#inp", "type": "string"},
            {"id": "http://example2.com/#inp2", "type": "string"},
        ],
        "outputs": [
            {
                "id": "http://example2.com/#out",
                "type": "string",
                "source": "http://example2.com/#step2/out",
            }
        ],
        "steps": [
            {
                "id": "http://example2.com/#step1",
                "scatter": "http://example2.com/#step1/inp",
                "in": [
                    {
                        "id": "http://example2.com/#step1/inp",
                        "source": "http://example2.com/#inp",
                    },
                    {
                        "id": "http://example2.com/#step1/inp2",
                        "source": "http://example2.com/#inp2",
                    },
                    {
                        "id": "http://example2.com/#step1/inp3",
                        "source": [
                            "http://example2.com/#inp",
                            "http://example2.com/#inp2",
                        ],
                    },
                ],
                "out": ["http://example2.com/#step1/out"],
            },
            {
                "id": "http://example2.com/#step2",
                "scatter": "http://example2.com/#step2/inp",
                "in": [
                    {
                        "id": "http://example2.com/#step2/inp",
                        "source": "http://example2.com/#step1/out",
                    }
                ],
                "out": ["http://example2.com/#step2/out"],
            },
        ],
    } == ra


def test_examples() -> None:
    for a in ["field_name", "ident_res", "link_res", "vocab_res"]:
        path = get_data(f"metaschema/{a}_schema.yml")
        assert path
        ldr, _, _, _ = schema_salad.schema.load_schema(path)
        path2 = get_data(f"metaschema/{a}_src.yml")
        assert path2
        yaml = yaml_no_ts()
        with open(path2) as src_fp:
            src = ldr.resolve_all(yaml.load(src_fp), "", checklinks=False)[0]
        path3 = get_data(f"metaschema/{a}_proc.yml")
        assert path3
        with open(path3) as src_proc:
            proc = yaml.load(src_proc)
        assert proc == src


def test_yaml_float_test() -> None:
    assert yaml_no_ts().load("float-test: 2e-10")["float-test"] == 2e-10


def test_typedsl_ref() -> None:
    ldr = Loader({})
    ldr.add_context(
        {
            "File": "http://example.com/File",
            "null": "http://example.com/null",
            "array": "http://example.com/array",
            "type": {"@type": "@vocab", "typeDSL": True},
        }
    )

    ra, _ = ldr.resolve_all(cmap({"type": "File"}), "")
    assert {"type": "File"} == ra

    ra, _ = ldr.resolve_all(cmap({"type": "File?"}), "")
    assert {"type": ["null", "File"]} == ra

    ra, _ = ldr.resolve_all(cmap({"type": "File[]"}), "")
    assert {"type": {"items": "File", "type": "array"}} == ra

    ra, _ = ldr.resolve_all(cmap({"type": "File[]?"}), "")
    assert {"type": ["null", {"items": "File", "type": "array"}]} == ra


def test_secondaryFile_dsl_ref() -> None:
    ldr = Loader({})
    ldr.add_context({"secondaryFiles": {"secondaryFilesDSL": True}})

    ra, _ = ldr.resolve_all(cmap({"secondaryFiles": ".foo"}), "")
    assert {"secondaryFiles": {"pattern": ".foo", "required": None}} == ra

    ra, _ = ldr.resolve_all(cmap({"secondaryFiles": ".foo?"}), "")
    assert {"secondaryFiles": {"pattern": ".foo", "required": False}} == ra

    ra, _ = ldr.resolve_all(cmap({"secondaryFiles": [".foo"]}), "")
    assert {"secondaryFiles": [{"pattern": ".foo", "required": None}]} == ra

    ra, _ = ldr.resolve_all(cmap({"secondaryFiles": [".foo?"]}), "")
    assert {"secondaryFiles": [{"pattern": ".foo", "required": False}]} == ra


def test_scoped_id() -> None:
    ldr = Loader({})
    ctx = {
        "id": "@id",
        "location": {"@id": "@id", "@type": "@id"},
        "bar": "http://example.com/bar",
        "ex": "http://example.com/",
    }  # type: ContextType
    ldr.add_context(ctx)

    ra, _ = ldr.resolve_all(cmap({"id": "foo", "bar": {"id": "baz"}}), "http://example.com")
    assert {
        "id": "http://example.com/#foo",
        "bar": {"id": "http://example.com/#foo/baz"},
    } == ra

    g = makerdf(None, ra, ctx)
    g.serialize(destination=stdout(), format="n3")

    ra, _ = ldr.resolve_all(
        cmap({"location": "foo", "bar": {"location": "baz"}}),
        "http://example.com",
        checklinks=False,
    )
    assert {
        "location": "http://example.com/foo",
        "bar": {"location": "http://example.com/baz"},
    } == ra

    g = makerdf(None, ra, ctx)
    g.serialize(destination=stdout(), format="n3")

    ra, _ = ldr.resolve_all(
        cmap({"id": "foo", "bar": {"location": "baz"}}),
        "http://example.com",
        checklinks=False,
    )
    assert {
        "id": "http://example.com/#foo",
        "bar": {"location": "http://example.com/baz"},
    } == ra

    g = makerdf(None, ra, ctx)
    g.serialize(destination=stdout(), format="n3")

    ra, _ = ldr.resolve_all(
        cmap({"location": "foo", "bar": {"id": "baz"}}),
        "http://example.com",
        checklinks=False,
    )
    assert {
        "location": "http://example.com/foo",
        "bar": {"id": "http://example.com/#baz"},
    } == ra

    g = makerdf(None, ra, ctx)
    g.serialize(destination=stdout(), format="n3")


def test_rdf_datetime() -> None:
    """Affirm that datetime objects can be serialized in makerdf()."""
    ldr = Loader({})
    ctx: ContextType = {
        "id": "@id",
        "location": {"@id": "@id", "@type": "@id"},
        "bar": "http://example.com/bar",
        "ex": "http://example.com/",
    }
    ldr.add_context(ctx)

    ra: CommentedMap = cast(
        CommentedMap,
        ldr.resolve_all(
            cmap(
                {
                    "id": "foo",
                    "bar": {"id": "baz"},
                }
            ),
            "http://example.com",
        )[0],
    )
    ra["s:dateCreated"] = datetime.datetime(2020, 10, 8)

    g = makerdf(None, ra, ctx)
    g.serialize(destination=stdout(), format="n3")
    g2 = makerdf(None, CommentedSeq([ra]), ctx)
    g2.serialize(destination=stdout(), format="n3")


def test_yaml_datetime() -> None:
    """Affirm that yaml_no_ts prevents the creation of datetime objects."""
    example: Dict[str, Any] = {
        "id": "foo",
        "bar": {"id": "baz"},
    }
    example["s:dateCreated"] = datetime.datetime(2020, 10, 8)
    yaml = yaml_no_ts()
    stream = StringIO()
    yaml.dump(example, stream)
    stream2 = StringIO(stream.getvalue())
    example2 = yaml.load(stream2)
    assert isinstance(example2["s:dateCreated"], str)


def test_subscoped_id() -> None:
    ldr = Loader({})
    ctx = {
        "id": "@id",
        "bar": {"subscope": "bar"},
    }  # type: ContextType
    ldr.add_context(ctx)

    ra, _ = ldr.resolve_all(cmap({"id": "foo", "bar": {"id": "baz"}}), "http://example.com")
    assert {
        "id": "http://example.com/#foo",
        "bar": {"id": "http://example.com/#foo/bar/baz"},
    } == ra


def test_mixin() -> None:
    base_url = file_uri(os.path.join(os.getcwd(), "tests"))
    ldr = Loader({})
    path = get_data("tests/mixin.yml")
    assert path
    ra = ldr.resolve_ref(cmap({"$mixin": path, "one": "five"}), base_url=base_url)
    assert {"id": "four", "one": "five"} == ra[0]
    ldr = Loader({"id": "@id"})

    ra = ldr.resolve_all(
        cmap([{"id": "a", "m": {"$mixin": path}}, {"id": "b", "m": {"$mixin": path}}]),
        base_url=base_url,
    )
    assert [
        {"id": base_url + "#a", "m": {"id": base_url + "#a/four", "one": "two"}},
        {"id": base_url + "#b", "m": {"id": base_url + "#b/four", "one": "two"}},
    ] == ra[0]


def test_fragment() -> None:
    ldr = Loader({"id": "@id"})
    path = get_data("tests/frag.yml#foo2")
    assert path
    b = ldr.resolve_ref(path)[0]
    assert isinstance(b, CommentedMap)
    assert {"id": b["id"], "bar": "b2"} == b


def test_file_uri() -> None:
    # Note: this test probably won't pass on Windows.  Someone with a
    # windows box should add an alternate test.
    assert "file:///foo/bar%20baz/quux" == file_uri("/foo/bar baz/quux")
    assert os.path.normpath("/foo/bar baz/quux") == uri_file_path("file:///foo/bar%20baz/quux")
    assert "file:///foo/bar%20baz/quux%23zing%20zong" == file_uri("/foo/bar baz/quux#zing zong")
    assert "file:///foo/bar%20baz/quux#zing%20zong" == file_uri(
        "/foo/bar baz/quux#zing zong", split_frag=True
    )
    assert os.path.normpath("/foo/bar baz/quux#zing zong") == uri_file_path(
        "file:///foo/bar%20baz/quux#zing%20zong"
    )


def test_sourceline() -> None:
    ldr = Loader({"id": "@id"})
    path = get_data("tests/frag.yml")
    assert path
    b, _ = ldr.resolve_ref(path)

    class TestExp(Exception):
        pass

    try:
        with SourceLine(b, 1, TestExp, False):
            raise Exception("Whoops")
    except TestExp as e:
        assert str(e).endswith("frag.yml:3:3: Whoops"), e
    except Exception as exc:
        raise AssertionError("Unexpected exception" + str(exc)) from exc


def test_cmap() -> None:
    # Test bugfix that cmap won't fail when given a CommentedMap with no lc.data
    cmap(CommentedMap((("foo", "bar"), ("baz", ["quux"]))))
    cmap(CommentedSeq(("foo", [], "bar")))


def test_blank_node_id() -> None:
    # Test that blank nodes are passed through and not considered
    # relative paths.  Blank nodes (also called anonymous ids) are
    # URIs starting with "_:".  They are randomly generated
    # placeholders mainly used internally where an id is needed but
    # was not given.

    ldr = Loader({})
    ctx = {"id": "@id"}  # type: ContextType
    ldr.add_context(ctx)

    ra, _ = ldr.resolve_all(cmap({"id": "_:foo"}), "http://example.com")
    assert {"id": "_:foo"} == ra


def test_can_use_Any() -> None:
    """Test that 'type: Any' can be used"""
    path = get_data("tests/test_schema/cwltest-schema.yml")
    assert path
    (
        document_loader,
        avsc_names,
        schema_metadata,
        metaschema_loader,
    ) = schema_salad.schema.load_schema(path)


def test_nullable_links() -> None:
    ldr = schema_salad.ref_resolver.Loader({})
    ctx = {"link": {"@type": "@id"}}  # type: ContextType
    ldr.add_context(ctx)

    ra, _ = ldr.resolve_all(cmap({"link": None}), "http://example.com", checklinks=True)
    assert {"link": None} == ra



"""
Test ``schema-salad-doc``.

(also known as ``schema-salad-tool --print-doc``)

For convenience, tests are checking exact strings. In the event of changes in
the "mistune" package, makedoc.py, or other changes, feel free to modify the test
strings as long as the new HTML renders the same way in typical browsers.

Likewise, if the schema-salad metaschema changes and it is missing one or more
of the features tested below, then please copy those old features to a new file
and update the affected tests to use those new file(s).
"""
import hashlib
import inspect
import json
import tempfile
from io import StringIO
from pathlib import Path
from typing import Optional

import pytest

from schema_salad.makedoc import makedoc

from .util import get_data


def test_schema_salad_inherit_docs() -> None:
    """Test schema-salad-doc when types inherit and override values from parent types."""
    schema_path = get_data("tests/inherited-attributes.yml")
    assert schema_path
    stdout = StringIO()
    makedoc(stdout, schema_path)

    # The parent ID documentation (i.e. Parent ID) must appear exactly once.
    assert 1 == stdout.getvalue().count("Parent ID")


def generate_doc(schema_data: Optional[str] = None) -> str:
    """Avoid error when calling fixture directly."""
    stdout = StringIO()
    if schema_data:
        with tempfile.NamedTemporaryFile(mode="w", encoding="utf-8", suffix=".yml") as tmp_file:
            tmp_file.write(schema_data)
            tmp_file.flush()
            tmp_file.seek(0)
            makedoc(stdout, tmp_file.name)
    else:
        schema_path = get_data("metaschema/metaschema.yml")
        assert schema_path
        makedoc(stdout, schema_path)
    return stdout.getvalue()


@pytest.fixture(scope="session", name="metaschema_doc")
def fixture_metaschema_doc() -> str:
    """Pytest Fixture of the rendered HTML for the metaschema schema."""
    return generate_doc()


def test_doc_fenced_code_contents_preserved() -> None:
    """
    Fenced code contents are not interpreted as Markdown definitions and converted into erroneous HTML.

    An example of problem case is when a definition looks like a Markdown list
    (e.g.: a YAML array). It must not be converted into HTML contents with list tags.
    However, special characters (e.g.: ``<``, ``>``) must still be escaped,
    otherwise they will not be correctly rendered within an HTML ``<pre><code>`` block.
    """
    data = inspect.cleandoc(
        """
        Option one generic example:
        ```
        some_cwl_field:
          - key_field: a_complex_type1
            field2: foo
            field3: bar
          - key_field: a_complex_type2
            field2: foo2
            field3: bar2
          - key_field: a_complex_type3
        ```

        Option one generic example:
        ```
        inputs:
          - id: workflow_input01
            type: string
          - id: workflow_input02
            type: File
            format: http://edamontology.org/format_2572
        ```

        Special Characters:
        ```
        data:
          - test: value 1 < 2 is true
          - test: value 2 > 1 is false
        ```
        """
    )
    schema_data = json.dumps(
        [
            {
                "name": "https://w3id.org/cwl/salad#test",
                "type": "documentation",
                "doc": data,
            }
        ]
    )
    schema_doc = generate_doc(schema_data)
    schema_doc = schema_doc.split("</head>")[-1]  # just for nicer/readable assert diffs
    assert (
        "<p>Option one generic example:</p>\n"
        "<pre><code>some_cwl_field:\n"
        "  - key_field: a_complex_type1\n"
        "    field2: foo\n"
        "    field3: bar\n"
        "  - key_field: a_complex_type2\n"
        "    field2: foo2\n"
        "    field3: bar2\n"
        "  - key_field: a_complex_type3\n"
        "</code></pre>\n"
    ) in schema_doc
    assert (
        "<p>Option one generic example:</p>\n"
        "<pre><code>inputs:\n"
        "  - id: workflow_input01\n"
        "    type: string\n"
        "  - id: workflow_input02\n"
        "    type: File\n"
        "    format: http://edamontology.org/format_2572\n"
        "</code></pre>\n"
    ) in schema_doc
    assert (
        "<p>Special Characters:</p>\n"
        "<pre><code>data:\n"
        "  - test: value 1 &lt; 2 is true\n"
        "  - test: value 2 &gt; 1 is false\n"
        "</code></pre>\n"
    ) in schema_doc


def test_doc_headings_target_anchor(metaschema_doc: str) -> None:
    """Doc headers must have an id and section link."""
    assert (
        '<h1 id="Abstract" class="section">Abstract '
        '<a href="#Abstract">&sect;</a></h1>' in metaschema_doc
    )


def test_doc_render_table_of_contents(metaschema_doc: str) -> None:
    """The special Table of Contents token must be replaced with a rendered table."""
    assert "!--ToC--" not in metaschema_doc


def test_plain_links_autolinked(metaschema_doc: str) -> None:
    """Plan links should be treated as if they were wrapped in angle brackets."""
    assert (
        "This document is the product of the "
        '<a href="https://groups.google.com/forum/#!forum/common-workflow-language">'
        "Common Workflow Language working\ngroup</a>" in metaschema_doc
    )


def test_embedded_html_unescaped() -> None:
    """Raw HTML shouldn't get escaped."""
    schema_path = get_data("tests/inherited-attributes.yml")
    assert schema_path
    stdout = StringIO()
    makedoc(stdout, schema_path)
    html = stdout.getvalue()

    assert '<table class="table">' in html
    assert "&lt;table class=&quot;table&quot;&gt;" not in html


def test_multiline_list_entries_word_spacing(metaschema_doc: str) -> None:
    """Hanging indents in Markdown lists don't lead to wordsmushing."""
    assert "as itis poorly documented" not in metaschema_doc
    assert "base URI for the document used toresolve relative" not in metaschema_doc
    assert "The keys ofthe object are namespace prefixes" not in metaschema_doc
    assert "This field may list URIreferences to documents in RDF-XML" not in metaschema_doc
    assert "defines valid fields thatmake up a record type" not in metaschema_doc
    assert "set of symbols that arevalid value" not in metaschema_doc


def test_multiline_list_entries_without_indention(metaschema_doc: str) -> None:
    """Hanging indents are not required in Markdown lists."""
    # See https://daringfireball.net/projects/markdown/syntax#list
    # and https://spec.commonmark.org/0.30/#example-290
    # Some newlines in markdown are replaced by spaces purposely
    # to avoid invalid Markdown to HTML conversion as preserve words separated.
    assert (
        "<li><p>At least one record definition object which defines valid fields that "  # purposely no \n
        "make up a record type.  Record field definitions include the valid types\n"
        "that may be assigned to each field and annotations to indicate fields\n"
        'that represent identifiers and links, described below in "Semantic\n'
        'Annotations".</p>\n'
        "</li>\n"
        "<li><p>Any number of enumerated type objects which define a set of finite "
        "set of symbols that are "  # purposely no \n
        "valid value of the type.</p>\n"
        "</li>\n"
        "<li><p>Any number of documentation objects which allow in-line "
        "documentation of the schema.</p>\n"
        "</li>" in metaschema_doc
    )
    assert (
        "<li>At least one record definition object which defines valid fields that</li>"
        not in metaschema_doc
    )


def test_detect_changes_in_html(metaschema_doc: str, tmp_path: Path) -> None:
    """Catch all for changes in HTML output, please adjust if the changes are innocent."""
    # If the hash changed because the metaschema itself changed (without changes
    # to makedoc.py or the version of mistune) then you can directly update the
    # hash value below.
    #
    # Otherwise, follow this procedure verify that the changed HTML rendering
    # is acceptable (or use make 'check-metaschema-diff' and 'compute-metaschema-hash'):
    # 1. Render the metaschema schema into using an older, known-working version
    #    of schema-salad:
    #    `schema-salad-doc schema_salad/metaschema/metaschema.yml > /tmp/metaschema.orig.html`
    # 2. Render the metaschema schema into using proposed changed codebase
    #    `schema-salad-doc schema_salad/metaschema/metaschema.yml > /tmp/metaschema.new.html`
    # 3. Confirm the other tests in this file pass using the new code/mistune,
    #    adjusting the test strings if the changes are truly innocent.
    # 4. Check the `diff` between the saved HTML pages to check for obvious problems
    #    `diff /tmp/metaschema.orig.html /tmp/metaschema.new.html`
    # 5. Check the HTML in both Firefox and Chrome, especially near areas
    #    of differences in the diff
    # 6. If the changes are agreeable, then update the hash below
    hasher = hashlib.sha256()
    hasher.update(metaschema_doc.encode("utf-8"))
    result = tmp_path / "result.html"
    with open(result, "w") as h:
        h.write(metaschema_doc)
    assert (
        hasher.hexdigest() == "5fd520208b242c0fbf50235c83c88bc462f2b8d67b3dfbb9ffb399d7851fdab2"
    ), result



from pathlib import Path
from typing import Any, Dict, List, cast

from schema_salad import schema

from .util import get_data_uri

cwl_file_uri = get_data_uri("tests/test_schema/CommonWorkflowLanguage.yml")


def test_extend_and_specialize_enums(tmp_path: Path) -> None:
    document_loader, _, _, metaschema_loader = schema.load_schema(cwl_file_uri)
    schema_raw_doc = metaschema_loader.fetch(cwl_file_uri)
    schema_doc, _ = metaschema_loader.resolve_all(schema_raw_doc, cwl_file_uri)

    j = schema.extend_and_specialize(cast(List[Dict[str, Any]], schema_doc), document_loader)
    CWLType = next((x for x in j if x["name"] == "https://w3id.org/cwl/cwl#CWLType"), None)
    assert CWLType is not None
    symbols = [
        "https://w3id.org/cwl/salad#null",
        "http://www.w3.org/2001/XMLSchema#boolean",
        "http://www.w3.org/2001/XMLSchema#int",
        "http://www.w3.org/2001/XMLSchema#long",
        "http://www.w3.org/2001/XMLSchema#float",
        "http://www.w3.org/2001/XMLSchema#double",
        "http://www.w3.org/2001/XMLSchema#string",
        "https://w3id.org/cwl/cwl#File",
        "https://w3id.org/cwl/cwl#Directory",
    ]
    for symbol in symbols:
        assert symbol in CWLType["symbols"]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Modifications copyright (C) 2017-2018 Common Workflow Language.
"""
Contains the Schema classes.

A schema may be one of:
  A record, mapping field names to field value data;
  An enum, containing one of a small set of symbols;
  An array of values, all of the same schema;
  A union of other schemas;
  A unicode string;
  A 32-bit signed int;
  A 64-bit signed long;
  A 32-bit floating-point float;
  A 64-bit floating-point double;
  A boolean; or
  Null.
"""
from typing import Any, Dict, List, Optional, Tuple, Union, cast

from mypy_extensions import mypyc_attr

from schema_salad.exceptions import SchemaException

#
# Constants
#

PRIMITIVE_TYPES = ("null", "boolean", "string", "int", "long", "float", "double")

NAMED_TYPES = ("enum", "record")

VALID_TYPES = PRIMITIVE_TYPES + NAMED_TYPES + ("array", "union")

SCHEMA_RESERVED_PROPS = (
    "type",
    "name",
    "namespace",
    "fields",  # Record
    "items",  # Array
    "symbols",  # Enum
    "doc",
)

# need recursion support in mypy/mypyc for a comprehensive JSON type
# MappingDataType = Dict[str, Union[PropType, List[PropsType]]]
# was: Union[str, MappingDataType, List[MappingDataType]]
JsonDataType = Any
AtomicPropType = Union[None, str, int, float, bool, "Schema", List[str], List["Field"]]
PropType = Union[
    AtomicPropType,
    Dict[str, AtomicPropType],
    List[Dict[str, AtomicPropType]],
]
PropsType = Dict[str, PropType]

FIELD_RESERVED_PROPS = ("default", "name", "doc", "order", "type")

VALID_FIELD_SORT_ORDERS = ("ascending", "descending", "ignore")

#
# Exceptions
#


class AvroException(SchemaException):
    pass


class SchemaParseException(AvroException):
    pass


#
# Base Classes
#


@mypyc_attr(serializable=True)
class Schema:
    """Base class for all Schema classes."""

    def __init__(self, atype: str, other_props: Optional[PropsType] = None) -> None:
        # Ensure valid ctor args
        if not isinstance(atype, str):
            raise SchemaParseException(
                f"Schema type {atype!r} must be a string, was {type(atype)!r}."
            )
        elif atype not in VALID_TYPES:
            fail_msg = f"{atype} is not a valid type."
            raise SchemaParseException(fail_msg)

        # add members
        if not hasattr(self, "_props"):
            self._props: PropsType = {}
        self.set_prop("type", atype)
        self.type = atype
        self._props.update(other_props or {})

    # Read-only properties dict. Printing schemas
    # creates JSON properties directly from this dict.
    @property
    def props(self) -> PropsType:
        return self._props

    # utility functions to manipulate properties dict
    def get_prop(self, key: str) -> Optional[PropType]:
        return self._props.get(key)

    def set_prop(self, key: str, value: Optional[PropType]) -> None:
        self._props[key] = value


class Name:
    """Class to describe Avro name."""

    def __init__(
        self,
        name_attr: Optional[str] = None,
        space_attr: Optional[str] = None,
        default_space: Optional[str] = None,
    ) -> None:
        """
        Formulate full name according to the specification.

        @arg name_attr: name value read in schema or None.
        @arg space_attr: namespace value read in schema or None.
        @ard default_space: the current default space or None.
        """
        # Ensure valid ctor args

        def validate(val: Optional[str], name: str) -> None:
            if (isinstance(val, str) and val != "") or val is None:
                # OK
                return
            fail_msg = f"{name} must be non-empty string or None."
            raise SchemaParseException(fail_msg)

        validate(name_attr, "Name")
        validate(space_attr, "Space")
        validate(default_space, "Default space")

        self._full: Optional[str] = name_attr

        if name_attr is None or name_attr == "":
            return

        if name_attr.find(".") < 0:
            if (space_attr is not None) and (space_attr != ""):
                self._full = f"{space_attr}.{name_attr}"
            else:
                if (default_space is not None) and (default_space != ""):
                    self._full = f"{default_space}.{name_attr}"

    @property
    def fullname(self) -> Optional[str]:
        return self._full

    def get_space(self) -> Optional[str]:
        """Back out a namespace from full name."""
        if self._full is None:
            return None

        if self._full.find(".") > 0:
            return self._full.rsplit(".", 1)[0]
        else:
            return None


class Names:
    """Track name set and default namespace during parsing."""

    def __init__(self, default_namespace: Optional[str] = None) -> None:
        self.names: Dict[str, NamedSchema] = {}
        self.default_namespace = default_namespace

    def has_name(self, name_attr: str, space_attr: Optional[str]) -> bool:
        test = Name(name_attr, space_attr, self.default_namespace).fullname
        return test in self.names

    def get_name(self, name_attr: str, space_attr: Optional[str]) -> Optional["NamedSchema"]:
        """Fetch the stored schema for the given namespace."""
        test = Name(name_attr, space_attr, self.default_namespace).fullname
        if test not in self.names:
            return None
        return self.names[test]

    def add_name(
        self, name_attr: str, space_attr: Optional[str], new_schema: "NamedSchema"
    ) -> Name:
        """
        Add a new schema object to the name set.

        :param name_attr: name value read in schema
        :param space_attr: namespace value read in schema.

        :returns: the Name that was just added.
        """
        to_add = Name(name_attr, space_attr, self.default_namespace)

        if to_add.fullname in VALID_TYPES:
            fail_msg = f"{to_add.fullname} is a reserved type name."
            raise SchemaParseException(fail_msg)
        elif to_add.fullname in self.names:
            fail_msg = f"The name {to_add.fullname!r} is already in use."
            raise SchemaParseException(fail_msg)
        elif to_add.fullname is None:
            fail_msg = f"{to_add.fullname} is missing, but this is impossible."
            raise SchemaParseException(fail_msg)

        self.names[to_add.fullname] = new_schema
        return to_add


class NamedSchema(Schema):
    """Named Schemas specified in NAMED_TYPES."""

    def __init__(
        self,
        atype: str,
        name: str,
        namespace: Optional[str] = None,
        names: Optional[Names] = None,
        other_props: Optional[PropsType] = None,
    ) -> None:
        # Ensure valid ctor args
        if not name:
            raise SchemaParseException("Named Schemas must have a non-empty name.")
        elif not isinstance(name, str):
            raise SchemaParseException("The name property must be a string.")
        elif namespace is not None and not isinstance(namespace, str):
            raise SchemaParseException("The namespace property must be a string.")
        if names is None:
            raise SchemaParseException("Must provide Names.")

        # Call parent ctor
        Schema.__init__(self, atype, other_props)

        # Add class members
        new_name = names.add_name(name, namespace, self)

        # Store name and namespace as they were read in origin schema
        self.set_prop("name", name)
        if namespace is not None:
            self.set_prop("namespace", new_name.get_space())

        # Store full name as calculated from name, namespace
        self._fullname = new_name.fullname

    # read-only properties
    @property
    def name(self) -> str:
        return cast(str, self.get_prop("name"))


@mypyc_attr(serializable=True)
class Field:
    def __init__(
        self,
        atype: JsonDataType,
        name: str,
        has_default: bool,
        default: Optional[Any] = None,
        order: Optional[str] = None,
        names: Optional[Names] = None,
        doc: Optional[Union[str, List[str]]] = None,
        other_props: Optional[PropsType] = None,
    ) -> None:
        # Ensure valid ctor args
        if not name:
            fail_msg = "Fields must have a non-empty name."
            raise SchemaParseException(fail_msg)
        elif not isinstance(name, str):
            fail_msg = "The name property must be a string."  # type: ignore[unreachable]
            raise SchemaParseException(fail_msg)
        elif order is not None and order not in VALID_FIELD_SORT_ORDERS:
            fail_msg = f"The order property {order} is not valid."
            raise SchemaParseException(fail_msg)

        # add members
        self._props: PropsType = {}
        self._has_default = has_default
        self._props.update(other_props or {})

        if isinstance(atype, str) and names is not None and names.has_name(atype, None):
            type_schema: Schema = cast(NamedSchema, names.get_name(atype, None))
        else:
            try:
                type_schema = make_avsc_object(atype, names)
            except Exception as e:
                raise SchemaParseException(
                    f"Type property {atype!r} not a valid Avro schema: {e}"
                ) from e
        self.set_prop("type", type_schema)
        self.set_prop("name", name)
        self.type = type_schema
        self.name = name
        # TODO(hammer): check to ensure default is valid
        if has_default:
            self.set_prop("default", default)
        if order is not None:
            self.set_prop("order", order)
        if doc is not None:
            self.set_prop("doc", doc)

    # read-only properties
    @property
    def default(self) -> Optional[Any]:
        return self.get_prop("default")

    # utility functions to manipulate properties dict
    def get_prop(self, key: str) -> Optional[PropType]:
        return self._props.get(key)

    def set_prop(self, key: str, value: Optional[PropType]) -> None:
        self._props[key] = value


#
# Primitive Types
#
class PrimitiveSchema(Schema):
    """Valid primitive types are in PRIMITIVE_TYPES."""

    def __init__(self, atype: str, other_props: Optional[PropsType] = None) -> None:
        # Ensure valid ctor args
        if atype not in PRIMITIVE_TYPES:
            raise AvroException(f"{atype} is not a valid primitive type.")

        # Call parent ctor
        Schema.__init__(self, atype, other_props=other_props)

        self.fullname = atype


#
# Complex Types (non-recursive)
#


class EnumSchema(NamedSchema):
    def __init__(
        self,
        name: str,
        namespace: Optional[str],
        symbols: List[str],
        names: Optional[Names] = None,
        doc: Optional[Union[str, List[str]]] = None,
        other_props: Optional[PropsType] = None,
    ) -> None:
        # Ensure valid ctor args
        if not isinstance(symbols, list):
            raise AvroException("Enum Schema requires a JSON array for the symbols property.")
        elif False in [isinstance(s, str) for s in symbols]:
            raise AvroException("Enum Schema requires all symbols to be JSON strings.")
        elif len(set(symbols)) < len(symbols):
            raise AvroException(f"Duplicate symbol: {symbols}")

        # Call parent ctor
        NamedSchema.__init__(self, "enum", name, namespace, names, other_props)

        # Add class members
        self.set_prop("symbols", symbols)
        if doc is not None:
            self.set_prop("doc", doc)

    # read-only properties
    @property
    def symbols(self) -> List[str]:
        return cast(List[str], self.get_prop("symbols"))


#
# Complex Types (recursive)
#


class ArraySchema(Schema):
    def __init__(
        self, items: JsonDataType, names: Names, other_props: Optional[PropsType] = None
    ) -> None:
        # Call parent ctor
        Schema.__init__(self, "array", other_props)
        # Add class members

        if names is None:
            raise SchemaParseException("Must provide Names.")
        if isinstance(items, str) and names.has_name(items, None):
            items_schema = cast(Schema, names.get_name(items, None))
        else:
            try:
                items_schema = make_avsc_object(items, names)
            except Exception as err:
                raise SchemaParseException(
                    f"Items schema ({items}) not a valid Avro schema: {err}. "
                    f"Known names: {list(names.names.keys())})."
                ) from err

        self.set_prop("items", items_schema)

    # read-only properties
    @property
    def items(self) -> Schema:
        return cast(Schema, self.get_prop("items"))


class UnionSchema(Schema):
    def __init__(
        self,
        schemas: List[JsonDataType],
        names: Names,
    ) -> None:
        """
        Initialize a new UnionSchema.

        :param names: a dictionary of schema objects
        """
        # Ensure valid ctor args
        if names is None:
            raise SchemaParseException("Must provide Names.")
        if not isinstance(schemas, list):
            raise SchemaParseException("Union schema requires a list of schemas.")

        # Call parent ctor
        Schema.__init__(self, "union")

        # Add class members
        schema_objects: List[Schema] = []
        for schema in schemas:
            if isinstance(schema, str) and names.has_name(schema, None):
                new_schema = cast(Schema, names.get_name(schema, None))
            else:
                try:
                    new_schema = make_avsc_object(schema, names)
                except Exception as err:
                    raise SchemaParseException(
                        f"Union item must be a valid Avro schema: {err}; {schema},"
                    ) from err
            # check the new schema
            if (
                new_schema.type in VALID_TYPES
                and new_schema.type not in NAMED_TYPES
                and new_schema.type in [schema.type for schema in schema_objects]
            ):
                raise SchemaParseException(f"{new_schema.type} type already in Union")
            elif new_schema.type == "union":
                raise SchemaParseException("Unions cannot contain other unions.")
            else:
                schema_objects.append(new_schema)
        self._schemas = schema_objects

    # read-only properties
    @property
    def schemas(self) -> List[Schema]:
        return self._schemas


class RecordSchema(NamedSchema):
    @staticmethod
    def make_field_objects(field_data: List[PropsType], names: Names) -> List[Field]:
        """We're going to need to make message parameters too."""
        field_objects: List[Field] = []
        parsed_fields: Dict[str, PropsType] = {}
        for field in field_data:
            if hasattr(field, "get") and callable(field.get):
                atype = field.get("type")
                name = cast(str, field.get("name"))

                # null values can have a default value of None
                has_default = False
                default = None
                if "default" in field:
                    has_default = True
                    default = field.get("default")
                order = field.get("order")
                if not (order is None or isinstance(order, str)):
                    raise SchemaParseException('"order" must be a string or None')
                doc = field.get("doc")
                if not (doc is None or isinstance(doc, str) or isinstance(doc, list)):
                    raise SchemaParseException('"doc" must be a string, list of strings, or None')
                else:
                    doc = cast(Union[str, List[str], None], doc)
                other_props = get_other_props(field, FIELD_RESERVED_PROPS)
                new_field = Field(atype, name, has_default, default, order, names, doc, other_props)
                parsed_fields[new_field.name] = field
            else:
                raise SchemaParseException(f"Not a valid field: {field}")
            field_objects.append(new_field)
        return field_objects

    def __init__(
        self,
        name: str,
        namespace: Optional[str],
        fields: List[PropsType],
        names: Names,
        schema_type: str = "record",
        doc: Optional[Union[str, List[str]]] = None,
        other_props: Optional[PropsType] = None,
    ) -> None:
        # Ensure valid ctor args
        if not isinstance(fields, list):
            raise SchemaParseException("Fields property must be a list of Avro schemas.")

        # Call parent ctor (adds own name to namespace, too)
        NamedSchema.__init__(self, schema_type, name, namespace, names, other_props)

        if schema_type == "record":
            old_default = names.default_namespace
            names.default_namespace = Name(name, namespace, names.default_namespace).get_space()

        # Add class members
        field_objects = RecordSchema.make_field_objects(fields, names)
        self.set_prop("fields", field_objects)
        if doc is not None:
            self.set_prop("doc", doc)

        if schema_type == "record":
            names.default_namespace = old_default

    # read-only properties
    @property
    def fields(self) -> List[Field]:
        return cast(List[Field], self.get_prop("fields"))


#
# Module Methods
#
def get_other_props(all_props: PropsType, reserved_props: Tuple[str, ...]) -> Optional[PropsType]:
    """
    Retrieve the non-reserved properties from a dictionary of properties.

    :param reserved_props: The set of reserved properties to exclude
    """
    if hasattr(all_props, "items") and callable(all_props.items):
        return {k: v for (k, v) in all_props.items() if k not in reserved_props}
    return None


def make_avsc_object(json_data: JsonDataType, names: Optional[Names] = None) -> Schema:
    """
    Build Avro Schema from data parsed out of JSON string.

    :param names: A Name object (tracks seen names and default space)
    """
    if names is None:
        names = Names()

    if isinstance(json_data, Dict) and json_data.get("name") == "org.w3id.cwl.salad.Any":
        del names.names["org.w3id.cwl.salad.Any"]
    elif not names.has_name("org.w3id.cwl.salad.Any", None):
        EnumSchema("org.w3id.cwl.salad.Any", None, ["Any"], names=names)

    # JSON object (non-union)
    if isinstance(json_data, dict):
        atype = json_data.get("type")
        other_props = get_other_props(json_data, SCHEMA_RESERVED_PROPS)
        if atype in PRIMITIVE_TYPES:
            primative_type = cast(str, atype)
            return PrimitiveSchema(primative_type, other_props)
        if atype in NAMED_TYPES:
            name = json_data.get("name")
            namespace = json_data.get("namespace", names.default_namespace)
            doc = json_data.get("doc")
            if not isinstance(name, str):
                raise SchemaParseException(f'"name" for type {atype} must be a string: {json_data}')
            if not (namespace is None or isinstance(namespace, str)):
                raise SchemaParseException(
                    f'"namespace" for type {atype} must be a string or None: {json_data}'
                )
            if not (doc is None or isinstance(doc, str) or isinstance(doc, list)):
                raise SchemaParseException(
                    f'"doc" for type {atype} must be a string, '
                    f"a list of strings, or None: {json_data}"
                )
            if atype == "enum":
                symbols = json_data.get("symbols")
                if not isinstance(symbols, list):
                    raise SchemaParseException(
                        f'"symbols" for type enum must be a list of strings: {json_data}'
                    )
                else:
                    symbols = cast(List[str], symbols)
                return EnumSchema(name, namespace, symbols, names, doc, other_props)
            if atype in ["record", "error"]:
                fields = json_data.get("fields", [])
                if not isinstance(fields, list):
                    raise SchemaParseException(
                        '"fields" for type {} must be a list of mappings: {}'.format(
                            atype, json_data
                        )
                    )
                else:
                    fields = cast(List[PropsType], fields)
                return RecordSchema(name, namespace, fields, names, atype, doc, other_props)
            raise SchemaParseException(f"Unknown Named Type: {atype}")
        if atype in VALID_TYPES:
            if atype == "array":
                items = json_data.get("items")
                return ArraySchema(items, names, other_props)
        if atype is None:
            raise SchemaParseException(f'No "type" property: {json_data}')
        raise SchemaParseException(f"Undefined type: {atype}")
    # JSON array (union)
    if isinstance(json_data, list):
        return UnionSchema(json_data, names)
    # JSON string (primitive)
    if json_data in PRIMITIVE_TYPES:
        return PrimitiveSchema(json_data)
    # not for us!
    fail_msg = f"Could not make an Avro Schema object from {json_data}."
    raise SchemaParseException(fail_msg)


def is_subtype(existing: PropType, new: PropType) -> bool:
    """Check if a new type specification is compatible with an existing type spec."""
    if existing == new:
        return True
    if isinstance(existing, list) and (new in existing):
        return True
    if existing == "Any":
        if new is None or new == [] or new == ["null"] or new == "null":
            return False
        if isinstance(new, list) and "null" in new:
            return False
        return True
    if (
        isinstance(existing, dict)
        and "type" in existing
        and existing["type"] == "array"
        and isinstance(new, dict)
        and "type" in new
        and new["type"] == "array"
    ):
        return is_subtype(existing["items"], new["items"])
    if (
        isinstance(existing, dict)
        and "type" in existing
        and existing["type"] == "enum"
        and isinstance(new, dict)
        and "type" in new
        and new["type"] == "enum"
    ):
        return is_subtype(existing["symbols"], new["symbols"])
    if (
        isinstance(existing, dict)
        and "type" in existing
        and existing["type"] == "record"
        and isinstance(new, dict)
        and "type" in new
        and new["type"] == "record"
    ):
        for new_field in cast(List[Dict[str, Any]], new["fields"]):
            new_field_missing = True
            for existing_field in cast(List[Dict[str, Any]], existing["fields"]):
                if new_field["name"] == existing_field["name"]:
                    if not is_subtype(existing_field["type"], new_field["type"]):
                        return False
                    new_field_missing = False
            if new_field_missing:
                return False
        return True
    if isinstance(existing, list) and isinstance(new, list):
        missing = False
        for _type in new:
            if _type not in existing and (not is_subtype(existing, cast(PropType, _type))):
                missing = True
        return not missing
    return False



import shutil
from pathlib import Path
from typing import Any, Dict, List, Optional, cast

from schema_salad import codegen, ref_resolver
from schema_salad.schema import load_schema

from .util import get_data


def test_cwl_gen(tmp_path: Path) -> None:
    topmed_example_path = get_data("tests/test_real_cwl/topmed/topmed_variant_calling_pipeline.cwl")
    assert topmed_example_path
    target_dir = tmp_path / "target"
    examples_dir = tmp_path / "examples"

    target_dir.mkdir()
    examples_dir.mkdir()
    shutil.copyfile(topmed_example_path, examples_dir / "valid_topmed.cwl")

    dotnet_codegen(cwl_file_uri, target_dir, examples=examples_dir)
    solution_path = target_dir / "Solution.sln"
    assert solution_path.exists()
    tests_dir = target_dir / "Test" / "src"
    assert tests_dir.exists()
    with open(tests_dir / "ExampleTests.cs") as f:
        assert "topmed" in f.read()


def test_meta_schema_gen(tmp_path: Path) -> None:
    target_dir = tmp_path / "target"
    target_dir.mkdir()
    dotnet_codegen(metaschema_file_uri, target_dir)
    solution_path = target_dir / "Solution.sln"
    assert solution_path.exists()
    src_dir = target_dir / "DotnetTest" / "src"
    assert src_dir.exists()
    record_schema_dir = src_dir / "RecordSchema.cs"
    assert record_schema_dir.exists()
    with open(record_schema_dir) as f:
        assert "public class RecordSchema : IRecordSchema, " "ISaveable\n{\n" in f.read()


def test_class_field(tmp_path: Path) -> None:
    schema_path = get_data_uri("tests/class_field_test.yml")
    assert schema_path
    target_dir = tmp_path / "target"

    target_dir.mkdir()
    dotnet_codegen(schema_path, target_dir)

    solution_path = target_dir / "Solution.sln"
    assert solution_path.exists()

    tests_dir = target_dir / "Test"
    assert tests_dir.exists()

    with open(target_dir / "DotnetTest" / "src" / "ClassFieldString.cs") as f:
        assert (
            'public ClassFieldString(string class_ = "ClassFieldString",'
            + " LoadingOptions? loadingOptions = null,"
            + " Dictionary<object, object>? extensionFields = null)\n    {\n"
            in f.read()
        )
    with open(target_dir / "DotnetTest" / "src" / "ClassFieldEnum.cs") as f:
        assert (
            "public ClassFieldEnum(ClassFieldEnum_class? class_ = null,"
            + " LoadingOptions? loadingOptions = null,"
            + " Dictionary<object, object>? extensionFields = null)\n    {\n"
            in f.read()
        )


def get_data_uri(resource_path: str) -> str:
    path = get_data(resource_path)
    assert path
    return ref_resolver.file_uri(path)


cwl_file_uri = get_data_uri("tests/test_schema/CommonWorkflowLanguage.yml")
metaschema_file_uri = get_data_uri("metaschema/metaschema.yml")


def dotnet_codegen(file_uri: str, target: Path, examples: Optional[Path] = None) -> None:
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(file_uri)
    schema_raw_doc = metaschema_loader.fetch(file_uri)
    schema_doc, schema_metadata = metaschema_loader.resolve_all(schema_raw_doc, file_uri)
    codegen.codegen(
        "dotnet",
        cast(List[Dict[str, Any]], schema_doc),
        schema_metadata,
        document_loader,
        package="DotnetTest",
        target=str(target),
        examples=str(examples) if examples else None,
    )



from typing import Iterable, Optional, Union

from mistune._types import *
from mistune.inline_parser import RendererT
from mistune.markdown import Markdown, ParseHook, RenderHook
from mistune.plugins import Plugin, PluginName
from mistune.renderers import BaseRenderer, DataT, HTMLRenderer, HTMLType
from typing_extensions import Literal

html: Markdown[HTMLType, HTMLRenderer]

RendererRef = Union[Literal["html", "ast"], BaseRenderer[DataT]]
PluginRef = Union[PluginName, Plugin]  # reference to register a plugin

def create_markdown(
    escape: bool = False,
    hard_wrap: bool = False,
    renderer: Optional[RendererRef[DataT]] = None,
    plugins: Optional[Iterable[PluginRef]] = None,
) -> Markdown[DataT, RendererT]: ...
def markdown(
    text: str,
    escape: bool = True,
    renderer: Optional[BaseRenderer[DataT]] = None,
    plugins: Optional[Iterable[PluginRef]] = None,
) -> str: ...

__version__: str



from typing import Any, Generic, Iterator, Optional, Type, TypeVar, overload

from rdflib.exceptions import Error

class PluginException(Error):
    pass

#: A generic type variable for plugins
PluginT = TypeVar("PluginT")

class Plugin(Generic[PluginT]):
    name: str
    kind: Type[PluginT]
    module_path: str
    class_name: str
    _class: Optional[Type[PluginT]]

    def __init__(
        self, name: str, kind: Type[PluginT], module_path: str, class_name: str
    ) -> None: ...
    def getClass(self) -> Type[PluginT]: ...

def register(name: str, kind: Type[Any], module_path: str, class_name: str) -> None: ...
def get(name: str, kind: Type[PluginT]) -> Type[PluginT]: ...
@overload
def plugins(name: Optional[str] = ..., kind: Type[PluginT] = ...) -> Iterator[Plugin[PluginT]]: ...
@overload
def plugins(name: Optional[str] = ..., kind: None = ...) -> Iterator[Plugin[Any]]: ...



"""Java code generator for a given schema salad definition."""
import os
import re
import shutil
import string
from io import StringIO
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    MutableMapping,
    MutableSequence,
    Optional,
    Set,
    Union,
)

import pkg_resources

from . import _logger, schema
from .codegen_base import CodeGenBase, TypeDef
from .exceptions import SchemaException
from .schema import shortname

# experiment at providing more typed objects building a optional type that allows
# referencing one or a list of objects. It is useful for improving the RootLoader
# for simple schema with a single root loader - but doesn't help with CWL at all and
# may even confuse things a bit so turning these off be default.
USE_ONE_OR_LIST_OF_TYPES = False

BASIC_JAVA_IDENTIFIER_RE = re.compile(r"[^0-9a-zA-Z]+")


def _ensure_directory_and_write(path: Path, contents: str) -> None:
    _safe_makedirs(path.parent)
    with open(path, mode="w", encoding="utf-8") as f:
        _logger.info("Writing file: %s", path)
        f.write(contents)


def doc_to_doc_string(doc: Optional[str], indent_level: int = 0) -> str:
    lead = " " + "  " * indent_level + "* " * indent_level
    if doc:
        doc_str = f"{lead}<BLOCKQUOTE>\n"
        doc_str += "\n".join([f"{lead}{line}" for line in doc.split("\n")])
        doc_str += f"{lead}</BLOCKQUOTE>"
    else:
        doc_str = ""
    return doc_str


def _safe_makedirs(path: Path) -> None:
    if not path.exists():
        os.makedirs(path)
        _logger.info("Created directory: %s", path)


_string_type_def = TypeDef(
    instance_type="String",
    init="new PrimitiveLoader<String>(String.class)",
    name="StringInstance",
    loader_type="Loader<String>",
)

_int_type_def = TypeDef(
    instance_type="Integer",
    init="new PrimitiveLoader<Integer>(Integer.class)",
    name="IntegerInstance",
    loader_type="Loader<Integer>",
)

_long_type_def = TypeDef(
    instance_type="Long",
    name="LongInstance",
    loader_type="Loader<Long>",
    init="new PrimitiveLoader<Long>(Long.class)",
)

_float_type_def = TypeDef(
    instance_type="Double",
    name="DoubleInstance",
    loader_type="Loader<Double>",
    init="new PrimitiveLoader<Double>(Double.class)",
)

_bool_type_def = TypeDef(
    instance_type="Boolean",
    name="BooleanInstance",
    loader_type="Loader<Boolean>",
    init="new PrimitiveLoader<Boolean>(Boolean.class)",
)

_null_type_def = TypeDef(
    instance_type="Object",
    name="NullInstance",
    loader_type="Loader<Object>",
    init="new NullLoader()",
)

_any_type_def = TypeDef(
    instance_type="Object",
    name="AnyInstance",
    init="new AnyLoader()",
    loader_type="Loader<Object>",
)

prims = {
    "http://www.w3.org/2001/XMLSchema#string": _string_type_def,
    "http://www.w3.org/2001/XMLSchema#int": _int_type_def,
    "http://www.w3.org/2001/XMLSchema#long": _long_type_def,
    "http://www.w3.org/2001/XMLSchema#float": _float_type_def,
    "http://www.w3.org/2001/XMLSchema#double": _float_type_def,
    "http://www.w3.org/2001/XMLSchema#boolean": _bool_type_def,
    "https://w3id.org/cwl/salad#null": _null_type_def,
    "https://w3id.org/cwl/salad#Any": _any_type_def,
    "string": _string_type_def,
    "int": _int_type_def,
    "long": _long_type_def,
    "float": _float_type_def,
    "double": _float_type_def,
    "boolean": _bool_type_def,
    "null": _null_type_def,
    "Any": _any_type_def,
}


class JavaCodeGen(CodeGenBase):
    def __init__(
        self,
        base: str,
        target: Optional[str],
        examples: Optional[str],
        package: str,
        copyright: Optional[str],
    ) -> None:
        super().__init__()
        self.base_uri = base
        self.examples = examples
        self.package = package
        self.artifact = self.package.split(".")[-1]
        self.copyright = copyright
        self.target_dir = Path(target or ".").resolve()
        rel_package_dir = self.package.replace(".", "/")
        self.rel_package_dir = Path(rel_package_dir)
        self.main_src_dir = self.target_dir / "src" / "main" / "java" / rel_package_dir
        self.test_src_dir = self.target_dir / "src" / "test" / "java" / rel_package_dir
        self.test_resources_dir = self.target_dir / "src" / "test" / "resources" / rel_package_dir

    def prologue(self) -> None:
        for src_dir in [self.main_src_dir, self.test_src_dir]:
            _safe_makedirs(src_dir)

        for primitive in prims.values():
            self.declare_type(primitive)

    @staticmethod
    def property_name(name: str) -> str:
        avn = schema.avro_field_name(name)
        return avn

    @staticmethod
    def safe_name(name: str) -> str:
        avn = JavaCodeGen.property_name(name)
        if avn in ("class", "extends", "abstract", "default", "package"):
            # reserved words
            avn = avn + "_"
        if avn and avn.startswith("anon."):
            avn = avn[5:]
        return avn

    def interface_name(self, n: str) -> str:
        return self.safe_name(n)

    def begin_class(
        self,
        classname: str,
        extends: MutableSequence[str],
        doc: str,
        abstract: bool,
        field_names: MutableSequence[str],
        idfield: str,
        optional_fields: Set[str],
    ) -> None:
        cls = self.interface_name(classname)
        self.current_class = cls
        self.current_class_is_abstract = abstract
        self.current_loader = StringIO()
        self.current_fieldtypes: Dict[str, TypeDef] = {}
        self.current_fields = StringIO()
        interface_doc_str = f"* Auto-generated interface for <I>{classname}</I><BR>"
        if not abstract:
            implemented_by = "This interface is implemented by {{@link {}Impl}}<BR>"
            interface_doc_str += implemented_by.format(cls)
        interface_doc_str += doc_to_doc_string(doc)
        class_doc_str = f"* Auto-generated class implementation for <I>{classname}</I><BR>"
        class_doc_str += doc_to_doc_string(doc)
        target = self.main_src_dir / f"{cls}.java"
        with open(target, "w") as f:
            _logger.info("Writing file: %s", target)
            if extends:
                ext = "extends " + ", ".join(self.interface_name(e) for e in extends) + ", Saveable"
            else:
                ext = "extends Saveable"
            f.write(
                """// Copyright Common Workflow Language project contributors
"""
            )
            if self.copyright:
                f.write(
                    """// {copyright}
""".format(
                        copyright=self.copyright
                    )
                )
            f.write(
                """//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package {package};

import {package}.utils.Saveable;

/**
{interface_doc_str}
 */
public interface {cls} {ext} {{""".format(
                    package=self.package,
                    cls=cls,
                    ext=ext,
                    interface_doc_str=interface_doc_str,
                )
            )

        if self.current_class_is_abstract:
            return

        target = self.main_src_dir / f"{cls}Impl.java"
        with open(target, "w") as f:
            _logger.info("Writing file: %s", target)
            f.write(
                """// Copyright Common Workflow Language project contributors
"""
            )
            if self.copyright:
                f.write(
                    """// {copyright}
""".format(
                        copyright=self.copyright
                    )
                )
            f.write(
                """//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package {package};

import {package}.utils.LoaderInstances;
import {package}.utils.LoadingOptions;
import {package}.utils.LoadingOptionsBuilder;
import {package}.utils.SaveableImpl;
import {package}.utils.ValidationException;

/**
{class_doc_str}
 */
public class {cls}Impl extends SaveableImpl implements {cls} {{
  private LoadingOptions loadingOptions_ = new LoadingOptionsBuilder().build();
  private java.util.Map<String, Object> extensionFields_ =
      new java.util.HashMap<String, Object>();
""".format(
                    package=self.package,
                    cls=cls,
                    class_doc_str=class_doc_str,
                )
            )
        self.current_loader.write(
            """
  /**
   * Used by {{@link {package}.utils.RootLoader}} to construct instances of {cls}Impl.
   *
   * @param __doc_            Document fragment to load this record object from (presumably a
                              {{@link java.util.Map}}).
   * @param __baseUri_        Base URI to generate child document IDs against.
   * @param __loadingOptions  Context for loading URIs and populating objects.
   * @param __docRoot_        ID at this position in the document (if available) (maybe?)
   * @throws ValidationException If the document fragment is not a {{@link java.util.Map}}
   *                             or validation of fields fails.
   */
  public {cls}Impl(
      final Object __doc_,
      final String __baseUri_,
      LoadingOptions __loadingOptions,
      final String __docRoot_) {{
    super(__doc_, __baseUri_, __loadingOptions, __docRoot_);
    // Prefix plumbing variables with '__' to reduce likelihood of collision with
    // generated names.
    String __baseUri = __baseUri_;
    String __docRoot = __docRoot_;
    if (!(__doc_ instanceof java.util.Map)) {{
      throw new ValidationException("{cls}Impl called on non-map");
    }}
    final java.util.Map<String, Object> __doc = (java.util.Map<String, Object>) __doc_;
    final java.util.List<ValidationException> __errors =
        new java.util.ArrayList<ValidationException>();
    if (__loadingOptions != null) {{
      this.loadingOptions_ = __loadingOptions;
    }}
""".format(
                cls=cls, package=self.package
            )
        )

    def end_class(self, classname: str, field_names: List[str]) -> None:
        """Finish this class."""
        with open(self.main_src_dir / f"{self.current_class}.java", "a") as f:
            f.write(
                """
}
"""
            )
        if self.current_class_is_abstract:
            return

        self.current_loader.write(
            """    if (!__errors.isEmpty()) {
      throw new ValidationException("Trying 'RecordField'", __errors);
    }
"""
        )
        for fieldname in field_names:
            fieldtype = self.current_fieldtypes.get(fieldname)
            if fieldtype is None:
                continue
            self.current_loader.write(
                """    this.{safename} = ({type}) {safename};
""".format(
                    safename=self.safe_name(fieldname), type=fieldtype.instance_type
                )
            )

        self.current_loader.write("""  }""")
        target = self.main_src_dir / f"{self.current_class}Impl.java"
        with open(
            target,
            "a",
        ) as f:
            f.write(self.current_fields.getvalue())
            f.write(self.current_loader.getvalue())
            f.write(
                """
}
"""
            )

    def type_loader(self, type_declaration: Union[List[Any], Dict[str, Any], str]) -> TypeDef:
        """Parse the given type declaration and declare its components."""
        if isinstance(type_declaration, MutableSequence):
            sub = [self.type_loader(i) for i in type_declaration]
            if len(sub) < 2:
                return sub[0]

            if len(sub) == 2:
                type_1 = sub[0]
                type_2 = sub[1]
                type_1_name = type_1.name
                type_2_name = type_2.name
                if type_1_name == "NullInstance" or type_2_name == "NullInstance":
                    non_null_type = type_1 if type_1.name != "NullInstance" else type_2
                    return self.declare_type(
                        TypeDef(
                            instance_type="java.util.Optional<{}>".format(
                                non_null_type.instance_type
                            ),
                            init=f"new OptionalLoader({non_null_type.name})",
                            name=f"optional_{non_null_type.name}",
                            loader_type="Loader<java.util.Optional<{}>>".format(
                                non_null_type.instance_type
                            ),
                        )
                    )
                if (
                    type_1_name == f"array_of_{type_2_name}"
                    or type_2_name == f"array_of_{type_1_name}"
                ) and USE_ONE_OR_LIST_OF_TYPES:
                    if type_1_name == f"array_of_{type_2_name}":
                        single_type = type_2
                        array_type = type_1
                    else:
                        single_type = type_1
                        array_type = type_2
                    fqclass = f"{self.package}.{single_type.instance_type}"
                    return self.declare_type(
                        TypeDef(
                            instance_type=f"{self.package}.utils.OneOrListOf<{fqclass}>",
                            init="new OneOrListOfLoader<{}>({}, {})".format(
                                fqclass, single_type.name, array_type.name
                            ),
                            name=f"one_or_array_of_{single_type.name}",
                            loader_type="Loader<{}.utils.OneOrListOf<{}>>".format(
                                self.package, fqclass
                            ),
                        )
                    )
            return self.declare_type(
                TypeDef(
                    instance_type="Object",
                    init="new UnionLoader(new Loader[] {{ {} }})".format(
                        ", ".join(s.name for s in sub)
                    ),
                    name="union_of_{}".format("_or_".join(s.name for s in sub)),
                    loader_type="Loader<Object>",
                )
            )
        if isinstance(type_declaration, MutableMapping):
            if type_declaration["type"] in (
                "array",
                "https://w3id.org/cwl/salad#array",
            ):
                i = self.type_loader(type_declaration["items"])
                instance_type = (
                    "java.util.List<String>"
                    if i.instance_type == "String"
                    else "java.util.List<Object>"
                )
                return self.declare_type(
                    TypeDef(
                        # special doesn't work out with subclassing, gotta be more clever
                        # instance_type="List<{}>".format(i.instance_type),
                        instance_type=instance_type,
                        name=f"array_of_{i.name}",
                        init=f"new ArrayLoader({i.name})",
                        loader_type=f"Loader<java.util.List<{i.instance_type}>>",
                    )
                )
            if type_declaration["type"] in ("enum", "https://w3id.org/cwl/salad#enum"):
                return self.type_loader_enum(type_declaration)
            if type_declaration["type"] in (
                "record",
                "https://w3id.org/cwl/salad#record",
            ):
                is_abstract = type_declaration.get("abstract", False)
                fqclass = "{}.{}".format(self.package, self.safe_name(type_declaration["name"]))
                return self.declare_type(
                    TypeDef(
                        instance_type=self.safe_name(type_declaration["name"]),
                        name=self.safe_name(type_declaration["name"]),
                        init="new RecordLoader<{clazz}>({clazz}{ext}.class)".format(
                            clazz=fqclass,
                            ext="Impl" if not is_abstract else "",
                        ),
                        loader_type=f"Loader<{fqclass}>",
                    )
                )
            raise SchemaException("wft {}".format(type_declaration["type"]))
        if type_declaration in prims:
            return prims[type_declaration]
        if type_declaration in ("Expression", "https://w3id.org/cwl/cwl#Expression"):
            return self.declare_type(
                TypeDef(
                    name=self.safe_name(type_declaration) + "Loader",
                    init="new ExpressionLoader()",
                    loader_type="Loader<String>",
                    instance_type="String",
                )
            )
        return self.collected_types[self.safe_name(type_declaration)]

    def type_loader_enum(self, type_declaration: Dict[str, Any]) -> TypeDef:
        symbols = [self.property_name(sym) for sym in type_declaration["symbols"]]
        for sym in symbols:
            self.add_vocab(shortname(sym), sym)
        clazz = self.safe_name(type_declaration["name"])
        symbols_decl = 'new String[] {{"{}"}}'.format('", "'.join(sym for sym in symbols))
        enum_path = self.main_src_dir / f"{clazz}.java"
        with open(enum_path, "w") as f:
            _logger.info("Writing file: %s", enum_path)
            f.write(
                """// Copyright Common Workflow Language project contributors
"""
            )
            if self.copyright:
                f.write(
                    """// {copyright}
""".format(
                        copyright=self.copyright
                    )
                )
            f.write(
                """//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package {package};

import {package}.utils.ValidationException;

public enum {clazz} {{
""".format(
                    package=self.package, clazz=clazz
                )
            )
            for i, sym in enumerate(symbols):
                suffix = "," if i < (len(symbols) - 1) else ";"
                const = self.safe_name(sym).replace("-", "_").replace(".", "_").upper()
                f.write(f"""  {const}("{sym}"){suffix}\n""")  # noqa: B907
            f.write(
                """
  private static String[] symbols = {symbols_decl};
  private String docVal;

  private {clazz}(final String docVal) {{
    this.docVal = docVal;
  }}

  public static {clazz} fromDocumentVal(final String docVal) {{
    for(final {clazz} val : {clazz}.values()) {{
      if(val.docVal.equals(docVal)) {{
        return val;
      }}
    }}
    throw new ValidationException(String.format("Expected one of %s", {clazz}.symbols, docVal));
  }}
}}
""".format(
                    clazz=clazz, symbols_decl=symbols_decl
                )
            )
        return self.declare_type(
            TypeDef(
                instance_type=clazz,
                name=self.safe_name(type_declaration["name"]),
                loader_type=f"Loader<{clazz}>",
                init=f"new EnumLoader({clazz}.class)",
            )
        )

    def declare_field(
        self,
        name: str,
        fieldtype: TypeDef,
        doc: Optional[str],
        optional: bool,
        subscope: str,
    ) -> None:
        fieldname = name
        property_name = self.property_name(fieldname)
        cap_case_property_name = property_name[0].upper() + property_name[1:]
        if cap_case_property_name == "Class":
            cap_case_property_name = "Class_"

        safename = self.safe_name(fieldname)
        self.current_fieldtypes[property_name] = fieldtype
        getter_doc_str = """  /**
   * Getter for property <I>{fieldname}</I><BR>
{field_doc_str}
   */
""".format(
            fieldname=fieldname, field_doc_str=doc_to_doc_string(doc, indent_level=1)
        )
        target = self.main_src_dir / f"{self.current_class}.java"
        with open(target, "a") as f:
            f.write(
                """
{getter_doc_str}
  {type} get{capfieldname}();""".format(
                    getter_doc_str=getter_doc_str,
                    capfieldname=cap_case_property_name,
                    type=fieldtype.instance_type,
                )
            )

        if self.current_class_is_abstract:
            return

        self.current_fields.write(
            """
  private {type} {safename};

{getter_doc_str}
  public {type} get{capfieldname}() {{
    return this.{safename};
  }}
""".format(
                safename=safename,
                capfieldname=cap_case_property_name,
                getter_doc_str=getter_doc_str,
                type=fieldtype.instance_type,
            )
        )

        self.current_loader.write(
            """    {type} {safename};
""".format(
                type=fieldtype.instance_type, safename=safename
            )
        )
        if optional:
            self.current_loader.write(
                """
    if (__doc.containsKey("{fieldname}")) {{
""".format(
                    fieldname=property_name
                )
            )
            spc = "  "
        else:
            spc = ""

        self.current_loader.write(
            """{spc}    try {{
{spc}      {safename} =
{spc}          LoaderInstances
{spc}              .{fieldtype}
{spc}              .loadField(__doc.get("{fieldname}"), __baseUri, __loadingOptions);
{spc}    }} catch (ValidationException e) {{
{spc}      {safename} = null; // won't be used but prevents compiler from complaining.
{spc}      final String __message = "the `{fieldname}` field is not valid because:";
{spc}      __errors.add(new ValidationException(__message, e));
{spc}    }}
""".format(
                fieldtype=fieldtype.name,
                safename=safename,
                fieldname=property_name,
                spc=spc,
            )
        )

        if optional:
            self.current_loader.write(
                """
    }} else {{
      {safename} = null;
    }}
""".format(
                    safename=safename
                )
            )

    def declare_id_field(
        self,
        name: str,
        fieldtype: TypeDef,
        doc: str,
        optional: bool,
    ) -> None:
        if self.current_class_is_abstract:
            return

        self.declare_field(name, fieldtype, doc, True, "")
        if optional:
            set_uri = """
    Boolean __original_is_null = {safename} == null;
    if ({safename} == null) {{
      if (__docRoot != null) {{
        {safename} = java.util.Optional.of(__docRoot);
      }} else {{
        {safename} = java.util.Optional.of("_:" + java.util.UUID.randomUUID().toString());
      }}
    }}
    if (__original_is_null) {{
        __baseUri = __baseUri_;
    }} else {{
        __baseUri = (String) {safename}.orElse(null);
    }}
"""
        else:
            set_uri = """
    if ({safename} == null) {{
      if (__docRoot != null) {{
        {safename} = __docRoot;
      }} else {{
        throw new ValidationException("Missing {fieldname}");
      }}
    }}
    __baseUri = (String) {safename};
"""

        self.current_loader.write(
            set_uri.format(safename=self.safe_name(name), fieldname=shortname(name))
        )

    def uri_loader(
        self,
        inner: TypeDef,
        scoped_id: bool,
        vocab_term: bool,
        ref_scope: Optional[int],
    ) -> TypeDef:
        instance_type = inner.instance_type or "Object"
        return self.declare_type(
            TypeDef(
                instance_type=instance_type,  # ?
                name=f"uri_{inner.name}_{scoped_id}_{vocab_term}_{ref_scope}",
                init="new UriLoader({}, {}, {}, {})".format(
                    inner.name,
                    self.to_java(scoped_id),
                    self.to_java(vocab_term),
                    self.to_java(ref_scope),
                ),
                is_uri=True,
                scoped_id=scoped_id,
                ref_scope=ref_scope,
                loader_type=f"Loader<{instance_type}>",
            )
        )

    def idmap_loader(
        self, field: str, inner: TypeDef, map_subject: str, map_predicate: Optional[str]
    ) -> TypeDef:
        instance_type = inner.instance_type or "Object"
        return self.declare_type(
            TypeDef(
                instance_type=instance_type,
                name=f"idmap_{self.safe_name(field)}_{inner.name}",
                init='new IdMapLoader({}, "{}", "{}")'.format(
                    inner.name, map_subject, map_predicate
                ),
                loader_type=f"Loader<{instance_type}>",
            )
        )

    def typedsl_loader(self, inner: TypeDef, ref_scope: Union[int, None]) -> TypeDef:
        instance_type = inner.instance_type or "Object"
        return self.declare_type(
            TypeDef(
                instance_type=instance_type,
                name=f"typedsl_{inner.name}_{ref_scope}",
                init=f"new TypeDslLoader({inner.name}, {ref_scope})",
                loader_type=f"Loader<{instance_type}>",
            )
        )

    def to_java(self, val: Any) -> Any:
        if val is True:
            return "true"
        elif val is None:
            return "null"
        elif val is False:
            return "false"
        return val

    def epilogue(self, root_loader: TypeDef) -> None:
        pd = "This project contains Java objects and utilities "
        pd = pd + ' auto-generated by <a href="https://github.com/'
        pd = pd + 'common-workflow-language/schema_salad">Schema Salad</a>'
        pd = pd + " for parsing documents corresponding to the "
        pd = pd + str(self.base_uri) + " schema."

        template_vars: MutableMapping[str, str] = dict(
            base_uri=self.base_uri,
            package=self.package,
            group_id=self.package,
            artifact_id=self.artifact,
            version="0.0.1-SNAPSHOT",
            project_name=self.package,
            project_description=pd,
            license_name="Apache License, Version 2.0",
            license_url="https://www.apache.org/licenses/LICENSE-2.0.txt",
        )

        def template_from_resource(resource: str) -> string.Template:
            template_str = pkg_resources.resource_string(__name__, f"java/{resource}").decode(
                "utf-8"
            )
            template = string.Template(template_str)
            return template

        def expand_resource_template_to(resource: str, path: Path) -> None:
            template = template_from_resource(resource)
            src = template.safe_substitute(template_vars)
            _ensure_directory_and_write(path, src)

        expand_resource_template_to("pom.xml", self.target_dir / "pom.xml")
        expand_resource_template_to("gitignore", self.target_dir / ".gitignore")
        expand_resource_template_to("package.html", self.main_src_dir / "package.html")
        expand_resource_template_to(
            "overview.html",
            self.target_dir / "src" / "main" / "javadoc" / "overview.html",
        )
        expand_resource_template_to(
            "MANIFEST.MF",
            self.target_dir / "src" / "main" / "resources" / "META-INF" / "MANIFEST.MF",
        )
        expand_resource_template_to("README.md", self.target_dir / "README.md")

        vocab = ""
        rvocab = ""
        for k in sorted(self.vocab.keys()):
            vocab += f"""    vocab.put("{k}", "{self.vocab[k]}");\n"""  # noqa: B907
            rvocab += f"""    rvocab.put("{self.vocab[k]}", "{k}");\n"""  # noqa: B907

        loader_instances = ""
        for _, collected_type in self.collected_types.items():
            loader_instances += "  public static {} {} = {};\n".format(
                collected_type.loader_type, collected_type.name, collected_type.init
            )

        example_tests = ""
        if self.examples:
            _safe_makedirs(self.test_resources_dir)
            utils_resources = self.test_resources_dir / "utils"
            if os.path.exists(utils_resources):
                shutil.rmtree(utils_resources)
            shutil.copytree(self.examples, utils_resources)
            for example_name in os.listdir(self.examples):
                if example_name.startswith("valid"):
                    basename = os.path.basename(example_name).rsplit(".", 1)[0]
                    basename = re.sub(BASIC_JAVA_IDENTIFIER_RE, "_", basename)
                    example_tests += """
  @org.junit.Test
  public void test{basename}ByString() throws Exception {{
    java.net.URL url = getClass().getResource("{example_name}");
    java.nio.file.Path resPath = java.nio.file.Paths.get(url.toURI());
    String yaml = new String(java.nio.file.Files.readAllBytes(resPath), "UTF8");
    RootLoader.loadDocument(yaml, url.toString());
  }}

  @org.junit.Test
  public void test{basename}ByPath() throws Exception {{
    java.net.URL url = getClass().getResource("{example_name}");
    java.nio.file.Path resPath = java.nio.file.Paths.get(url.toURI());
    RootLoader.loadDocument(resPath);
  }}

  @org.junit.Test
  public void test{basename}ByMap() throws Exception {{
    java.net.URL url = getClass().getResource("{example_name}");
    java.nio.file.Path resPath = java.nio.file.Paths.get(url.toURI());
    String yaml = new String(java.nio.file.Files.readAllBytes(resPath), "UTF8");
    java.util.Map<String, Object> doc;
    doc = (java.util.Map<String, Object>) YamlUtils.mapFromString(yaml);
    RootLoader.loadDocument(doc, url.toString());
  }}""".format(
                        basename=basename,
                        example_name=example_name,
                    )

        template_args: MutableMapping[str, str] = dict(
            package=self.package,
            vocab=vocab,
            rvocab=rvocab,
            loader_instances=loader_instances,
            root_loader_name=root_loader.name,
            root_loader_instance_type=root_loader.instance_type or "Object",
            example_tests=example_tests,
        )

        util_src_dirs = {
            "main_utils": self.main_src_dir,
            "test_utils": self.test_src_dir,
        }
        for util_src, util_target in util_src_dirs.items():
            for util in pkg_resources.resource_listdir(__name__, f"java/{util_src}"):
                src_path = util_target / "utils" / util
                src_template = template_from_resource(os.path.join(util_src, util))
                src = src_template.safe_substitute(template_args)
                _ensure_directory_and_write(src_path, src)

    def secondaryfilesdsl_loader(self, inner: TypeDef) -> TypeDef:
        instance_type = inner.instance_type or "Object"
        return self.declare_type(
            TypeDef(
                instance_type=instance_type,
                name=f"secondaryfilesdsl_{inner.name}",
                init=f"new SecondaryFilesDslLoader({inner.name})",
                loader_type=f"Loader<{instance_type}>",
            )
        )



"""Base class for the generation of loaders from schema-salad definitions."""
from collections import OrderedDict
from typing import Any, Dict, List, MutableSequence, Optional, Set, Union


class TypeDef:  # pylint: disable=too-few-public-methods
    """Schema Salad type description."""

    __slots__ = [
        "name",
        "init",
        "is_uri",
        "scoped_id",
        "ref_scope",
        "loader_type",
        "instance_type",
        "abstract",
    ]

    # switch to class-style typing.NamedTuple once support for Python < 3.6
    # is dropped
    def __init__(
        self,  # pylint: disable=too-many-arguments
        name: str,
        init: str,
        is_uri: bool = False,
        scoped_id: bool = False,
        ref_scope: Optional[int] = 0,
        loader_type: Optional[str] = None,
        instance_type: Optional[str] = None,
        abstract: bool = False,
    ) -> None:
        self.name = name
        self.init = init
        self.is_uri = is_uri
        self.scoped_id = scoped_id
        self.ref_scope = ref_scope
        self.abstract = abstract
        # Follow attributes used by Java but not Python.
        self.loader_type = loader_type
        self.instance_type = instance_type


class CodeGenBase:
    """Abstract base class for schema salad code generators."""

    def __init__(self) -> None:
        self.collected_types: OrderedDict[str, TypeDef] = OrderedDict()
        self.vocab: Dict[str, str] = {}

    def declare_type(self, declared_type: TypeDef) -> TypeDef:
        """Add this type to our collection, if needed."""
        if declared_type not in self.collected_types.values():
            self.collected_types[declared_type.name] = declared_type
        return declared_type

    def add_vocab(self, name: str, uri: str) -> None:
        """Add the given name as an abbreviation for the given URI."""
        self.vocab[name] = uri

    def prologue(self) -> None:
        """Trigger to generate the prolouge code."""
        raise NotImplementedError()

    @staticmethod
    def safe_name(name: str) -> str:
        """Generate a safe version of the given name."""
        raise NotImplementedError()

    def begin_class(
        self,  # pylint: disable=too-many-arguments
        classname: str,
        extends: MutableSequence[str],
        doc: str,
        abstract: bool,
        field_names: MutableSequence[str],
        idfield: str,
        optional_fields: Set[str],
    ) -> None:
        """Produce the header for the given class."""
        raise NotImplementedError()

    def end_class(self, classname: str, field_names: List[str]) -> None:
        """Signal that we are done with this class."""
        raise NotImplementedError()

    def type_loader(self, type_declaration: Union[List[Any], Dict[str, Any]]) -> TypeDef:
        """Parse the given type declaration and declare its components."""
        raise NotImplementedError()

    def declare_field(
        self,
        name: str,
        fieldtype: TypeDef,
        doc: Optional[str],
        optional: bool,
        subscope: str,
    ) -> None:
        """Output the code to load the given field."""
        raise NotImplementedError()

    def declare_id_field(
        self,
        name: str,
        fieldtype: TypeDef,
        doc: str,
        optional: bool,
    ) -> None:
        """Output the code to handle the given ID field."""
        raise NotImplementedError()

    def uri_loader(
        self,
        inner: TypeDef,
        scoped_id: bool,
        vocab_term: bool,
        ref_scope: Optional[int],
    ) -> TypeDef:
        """Construct the TypeDef for the given URI loader."""
        raise NotImplementedError()

    def idmap_loader(
        self, field: str, inner: TypeDef, map_subject: str, map_predicate: Optional[str]
    ) -> TypeDef:
        """Construct the TypeDef for the given mapped ID loader."""
        raise NotImplementedError()

    def typedsl_loader(self, inner: TypeDef, ref_scope: Optional[int]) -> TypeDef:
        """Construct the TypeDef for the given DSL loader."""
        raise NotImplementedError()

    def epilogue(self, root_loader: TypeDef) -> None:
        """Trigger to generate the epilouge code."""
        raise NotImplementedError()

    def secondaryfilesdsl_loader(self, inner: TypeDef) -> TypeDef:
        """Construct the TypeDef for secondary files."""
        raise NotImplementedError()



from typing import Dict, Optional, Union

from rdflib.graph import Graph

from schema_salad.avro.schema import Names
from schema_salad.schema import load_schema

from .util import get_data


def test_misc() -> None:
    path = get_data("tests/test_schema/no_field_schema.yml")
    assert path
    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(path)
    assert isinstance(avsc_names, Names)


def test_load_schema_cache() -> None:
    schemaid = "http://commonwl.org/schema_salad/test/schema.yml"

    path1 = get_data("tests/test_schema/misc_schema_v1.yml")
    assert path1

    with open(path1) as f:
        cache1: Optional[Dict[str, Union[str, Graph, bool]]] = {schemaid: f.read()}

    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(
        schemaid, cache=cache1
    )
    assert isinstance(avsc_names, Names)
    assert "EmptyType" not in document_loader.vocab

    path2 = get_data("tests/test_schema/misc_schema_v2.yml")
    assert path2

    with open(path2) as f:
        cache2: Optional[Dict[str, Union[str, Graph, bool]]] = {schemaid: f.read()}

    document_loader, avsc_names, schema_metadata, metaschema_loader = load_schema(
        schemaid, cache=cache2
    )
    assert isinstance(avsc_names, Names)
    assert "EmptyType" in document_loader.vocab



"""
Checks for accepting $schemas directive

run individually as py.test -k tests/test_schemas_directive.py
"""

import os
from typing import Any, Dict, Optional, Tuple, Union

from schema_salad.avro.schema import Names, SchemaParseException
from schema_salad.ref_resolver import Loader
from schema_salad.schema import load_and_validate, load_schema

from .util import get_data

test_dir_name = "tests/"


class TestSchemasDirective:
    """Ensure codegen-produced parsers accept $schemas directives"""

    document_loader: Loader
    avsc_names: Union[Names, SchemaParseException, None] = None
    schema_metadata: Optional[Dict[str, Any]] = None
    metaschema_loader: Optional[Loader] = None

    @classmethod
    def setup_class(cls) -> None:
        path = get_data("tests/test_schema/CommonWorkflowLanguage.yml")
        assert path
        (
            cls.document_loader,
            cls.avsc_names,
            schema_metadata,
            metaschema_loader,
        ) = load_schema(path)

    def load_cwl(self, src: str) -> Tuple[Any, Dict[str, Any]]:
        path = get_data(test_dir_name + src)
        assert path
        assert isinstance(self.avsc_names, Names)
        res = load_and_validate(self.document_loader, self.avsc_names, path, True)
        return res

    def test_dollarsign_schema(self) -> None:
        """EDAM.owl as a schema"""
        res = self.load_cwl(src="formattest2.cwl")

        # EDAM.owl resides in this directory
        assert os.path.split(str(res[0]["$schemas"][0]))[1] == "EDAM.owl"



#!/usr/bin/env python3

import os
import sys
from typing import List

from setuptools import setup

SETUP_DIR = os.path.dirname(__file__)
README = os.path.join(SETUP_DIR, "README.rst")

needs_pytest = {"pytest", "test", "ptr"}.intersection(sys.argv)
pytest_runner: List[str] = ["pytest < 8", "pytest-runner"] if needs_pytest else []

USE_MYPYC = False
# To compile with mypyc, a mypyc checkout must be present on the PYTHONPATH
if len(sys.argv) > 1 and sys.argv[1] == "--use-mypyc":
    sys.argv.pop(1)
    USE_MYPYC = True
if os.getenv("SCHEMA_SALAD_USE_MYPYC", None) == "1":
    USE_MYPYC = True

if USE_MYPYC and any(
    item in sys.argv
    for item in [
        "build",
        "bdist_wheel",
        "build_ext",
        "install",
        "install-lib",
        "bdist",
        "bdist_dumb",
        "bdist_rpm",
        "develop",
        "bdist_egg",
        "editable_wheel",
        "test",
    ]
):
    mypyc_targets = [
        # "schema_salad/codegen_base.py",  # interpreted classes cannot inherit from compiled
        # "schema_salad/codegen.py",
        # "schema_salad/exceptions.py", # leads to memory leaks
        "schema_salad/__init__.py",
        # "schema_salad/java_codegen.py",  # due to use of __name__
        "schema_salad/jsonld_context.py",
        "schema_salad/__main__.py",
        "schema_salad/main.py",
        "schema_salad/makedoc.py",
        # "schema_salad/python_codegen.py",  # due to use of __name__
        "schema_salad/ref_resolver.py",
        # "schema_salad/fetcher.py",  # to allow subclassing {Default,}Fetcher
        "schema_salad/schema.py",
        "schema_salad/sourceline.py",
        "schema_salad/utils.py",
        "schema_salad/validate.py",
        "schema_salad/avro/__init__.py",
        "schema_salad/avro/schema.py",
        # "schema_salad/tests/util.py",
        # "schema_salad/tests/test_print_oneline.py",
        # "schema_salad/tests/test_python_codegen.py",
        # "schema_salad/tests/test_java_codegen.py",
        # "schema_salad/tests/__init__.py",
        # "schema_salad/tests/test_cli_args.py",
        # "schema_salad/tests/test_fetch.py",
        # "schema_salad/tests/matcher.py",
        # "schema_salad/tests/test_cg.py",
        # "schema_salad/tests/test_examples.py",
        # "schema_salad/tests/test_errors.py",
        # "schema_salad/tests/test_real_cwl.py",
        # "schema_salad/tests/test_ref_resolver.py",
        # "schema_salad/tests/test_fp.py",
    ]

    from mypyc.build import mypycify

    opt_level = os.getenv("MYPYC_OPT_LEVEL", "3")
    ext_modules = mypycify(mypyc_targets, opt_level=opt_level, debug_level="0", verbose=True)
else:
    ext_modules = []

install_requires = [
    "requests >= 1.0",
    "ruamel.yaml >= 0.17.6, < 0.17.22;python_version>='3.7'",
    "ruamel.yaml >= 0.16.12, < 0.17.22",
    "rdflib >= 4.2.2, < 7.0.0",
    "rdflib-jsonld>=0.4.0, <= 0.6.1;python_version<='3.6'",
    "mistune>=2.0.3,<2.1",
    "CacheControl[filecache] >= 0.11.7, < 0.13",
    "mypy_extensions",
]

extras_require = {
    "docs": [
        "sphinx >= 2.2",
        "sphinx-rtd-theme",
        "pytest < 8",
        "sphinx-autoapi",
        "sphinx-autodoc-typehints",
        "typed_ast;python_version<'3.8'",
        "sphinxcontrib-autoprogram",
    ],
    "pycodegen": ["black"],
}

setup(
    name="schema-salad",
    version="8.4",  # update the VERSION prefix in the Makefile as well 🙂
    description="Schema Annotations for Linked Avro Data (SALAD)",
    long_description=open(README).read(),
    long_description_content_type="text/x-rst",
    author="Common workflow language working group",
    author_email="common-workflow-language@googlegroups.com",
    url="https://github.com/common-workflow-language/schema_salad",
    download_url="https://github.com/common-workflow-language/schema_salad/releases",
    ext_modules=ext_modules,
    license="Apache 2.0",
    python_requires=">=3.6,<3.12",
    setup_requires=pytest_runner + ["setuptools_scm"],
    packages=["schema_salad", "schema_salad.tests", "schema_salad.avro"],
    package_data={
        "schema_salad": [
            "metaschema/*",
            "py.typed",
            "dotnet/*",
            "dotnet/*/*",
            "dotnet/*/*/*",
            "java/*",
            "java/*/*",
            "typescript/*",
            "typescript/*/*",
            "typescript/*/*/*",
            "typescript/.*",
        ],
        "schema_salad.tests": [
            "*.json",
            "*.yml",
            "docimp/*",
            "*.owl",
            "*.cwl",
            "*.txt",
            "foreign/*.cwl",
            "test_real_cwl/*",
            "test_real_cwl/*/*",
            "test_schema/*",
        ],
    },
    install_requires=install_requires,
    extras_require=extras_require,
    test_suite="tests",
    tests_require=["pytest<8"],
    entry_points={
        "console_scripts": [
            "schema-salad-tool=schema_salad.main:main",
            "schema-salad-doc=schema_salad.makedoc:main",
        ]
    },
    zip_safe=True,
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Typing :: Typed",
    ],
    use_scm_version=True,
)



"""A schema language for describing JSON or YAML structured linked data documents."""

import logging
import sys
import warnings

__author__ = "peter.amstutz@curoverse.com"

_logger = logging.getLogger("salad")
_logger.addHandler(logging.StreamHandler())
_logger.setLevel(logging.INFO)

if sys.version_info < (3, 7):
    warnings.filterwarnings("ignore", message="urllib3.*", module="requests")



from typing import Dict, Protocol

from mistune.inline_parser import RendererT
from mistune.markdown import Markdown
from mistune.renderers import DataT
from typing_extensions import Literal

PluginName = Literal[
    "url",
    "strikethrough",
    "footnotes",
    "table",
    "task_lists",
    "def_list",
    "abbr",
]

class Plugin(Protocol):
    def __call__(self, markdown: Markdown[DataT, RendererT]) -> None: ...

PLUGINS: Dict[PluginName, Plugin]
